"""Legacy Indy Registry."""

import asyncio
import logging
import re
from asyncio import shield
from typing import List, Optional, Pattern, Sequence, Tuple

from anoncreds import (
    CredentialDefinition,
    RevocationRegistryDefinition,
    RevocationRegistryDefinitionPrivate,
)
from base58 import alphabet
from uuid_utils import uuid4

from ....anoncreds.default.legacy_indy.author import get_endorser_info
from ....cache.base import BaseCache
from ....config.injection_context import InjectionContext
from ....core.event_bus import EventBus
from ....core.profile import Profile, ProfileSession
from ....ledger.base import BaseLedger
from ....ledger.error import (
    LedgerError,
    LedgerObjectAlreadyExistsError,
    LedgerTransactionError,
)
from ....ledger.merkel_validation.constants import (
    GET_REVOC_REG_DELTA,
    GET_REVOC_REG_ENTRY,
    GET_SCHEMA,
)
from ....ledger.multiple_ledger.ledger_requests_executor import (
    GET_CRED_DEF,
    IndyLedgerRequestsExecutor,
)
from ....messaging.responder import BaseResponder
from ....multitenant.base import BaseMultitenantManager
from ....protocols.endorse_transaction.v1_0.manager import (
    TransactionManager,
    TransactionManagerError,
)
from ....protocols.endorse_transaction.v1_0.util import is_author_role
from ....revocation_anoncreds.models.issuer_cred_rev_record import IssuerCredRevRecord
from ....storage.error import StorageError
from ....utils import sentinel
from ....wallet.did_info import DIDInfo
from ...base import (
    AnonCredsObjectAlreadyExists,
    AnonCredsObjectNotFound,
    AnonCredsRegistrationError,
    AnonCredsResolutionError,
    AnonCredsSchemaAlreadyExists,
    BaseAnonCredsRegistrar,
    BaseAnonCredsResolver,
)
from ...events import RevListFinishedEvent
from ...issuer import CATEGORY_CRED_DEF, AnonCredsIssuer, AnonCredsIssuerError
from ...models.credential_definition import (
    CredDef,
    CredDefResult,
    CredDefState,
    CredDefValue,
    GetCredDefResult,
)
from ...models.revocation import (
    GetRevListResult,
    GetRevRegDefResult,
    RevList,
    RevListResult,
    RevListState,
    RevRegDef,
    RevRegDefResult,
    RevRegDefState,
    RevRegDefValue,
)
from ...models.schema import AnonCredsSchema, GetSchemaResult, SchemaResult, SchemaState
from ...models.schema_info import AnonCredsSchemaInfo
from ...revocation import (
    CATEGORY_REV_LIST,
    CATEGORY_REV_REG_DEF,
    CATEGORY_REV_REG_DEF_PRIVATE,
)
from .recover import generate_ledger_rrrecovery_txn

LOGGER = logging.getLogger(__name__)

# Defaults
DEFAULT_CRED_DEF_TAG = "default"
DEFAULT_SIGNATURE_TYPE = "CL"

# Common error messages
NO_LEDGER_AVAILABLE_MSG = "No ledger available"
MISSING_WALLET_TYPE_MSG = ": missing wallet-type?"
TRANSACTION_MANAGER_FAILED_MSG = "Transaction manager failed to create request: "
FAILED_TO_STORE_TRANSACTION_RECORD = "Failed to store transaction record"

# Common settings
ENDORSER_AUTO = "endorser.auto_request"
WALLET_TYPE = "wallet.type"


class LegacyIndyRegistry(BaseAnonCredsResolver, BaseAnonCredsRegistrar):
    """LegacyIndyRegistry."""

    def __init__(self):
        """Initialize an instance.

        Args:
            None

        """
        B58 = alphabet if isinstance(alphabet, str) else alphabet.decode("ascii")
        INDY_DID = rf"^(did:sov:)?[{B58}]{{21,22}}$"
        INDY_SCHEMA_ID = rf"^[{B58}]{{21,22}}:2:.+:[0-9.]+$"
        # the schema id can be just a number
        # (this is how the schema_id is referenced in a cred def)
        INDY_SCHEMA_TXN_ID = r"^[0-9.]+$"
        INDY_CRED_DEF_ID = (
            rf"^([{B58}]{{21,22}})"  # issuer DID
            f":3"  # cred def id marker
            f":CL"  # sig alg
            rf":(([1-9][0-9]*)|([{B58}]{{21,22}}:2:.+:[0-9.]+))"  # schema txn / id
            f":(.+)?$"  # tag
        )
        INDY_REV_REG_DEF_ID = (
            rf"^([{B58}]{{21,22}}):4:"
            rf"([{B58}]{{21,22}}):3:"
            rf"CL:(([1-9][0-9]*)|([{B58}]{{21,22}}:2:.+:[0-9.]+))(:.+)?:"
            rf"CL_ACCUM:(.+$)"
        )
        self._supported_identifiers_regex = re.compile(
            rf"{INDY_DID}|{INDY_SCHEMA_ID}|{INDY_SCHEMA_TXN_ID}|{INDY_CRED_DEF_ID}|{INDY_REV_REG_DEF_ID}"
        )

    @property
    def supported_identifiers_regex(self) -> Pattern:
        """Supported Identifiers Regular Expression."""
        return self._supported_identifiers_regex

    async def setup(self, context: InjectionContext) -> None:
        """Setup."""
        LOGGER.info("Successfully registered LegacyIndyRegistry")

    @staticmethod
    def make_schema_id(schema: AnonCredsSchema) -> str:
        """Derive the ID for a schema."""
        return f"{schema.issuer_id}:2:{schema.name}:{schema.version}"

    @staticmethod
    def make_cred_def_id(
        schema: GetSchemaResult,
        cred_def: CredDef,
    ) -> str:
        """Derive the ID for a credential definition."""
        signature_type = cred_def.type or DEFAULT_SIGNATURE_TYPE
        tag = cred_def.tag or DEFAULT_CRED_DEF_TAG

        try:
            seq_no = str(schema.schema_metadata["seqNo"])
        except KeyError as err:
            raise AnonCredsRegistrationError(
                "Legacy Indy only supports schemas from Legacy Indy"
            ) from err

        return f"{cred_def.issuer_id}:3:{signature_type}:{seq_no}:{tag}"

    @staticmethod
    def make_rev_reg_def_id(rev_reg_def: RevRegDef) -> str:
        """Derive the ID for a revocation registry definition."""
        return (
            f"{rev_reg_def.issuer_id}:4:{rev_reg_def.cred_def_id}:"
            f"{rev_reg_def.type}:{rev_reg_def.tag}"
        )

    async def get_schema(self, profile: Profile, schema_id: str) -> GetSchemaResult:
        """Get a schema from the registry."""

        multitenant_mgr = profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
        else:
            ledger_exec_inst = profile.inject(IndyLedgerRequestsExecutor)
        ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
            schema_id,
            txn_record_type=GET_SCHEMA,
        )

        if not ledger:
            reason = NO_LEDGER_AVAILABLE_MSG
            if not profile.settings.get_value(WALLET_TYPE):
                reason += MISSING_WALLET_TYPE_MSG
            raise AnonCredsResolutionError(reason)

        async with ledger:
            try:
                schema = await ledger.get_schema(schema_id)
                if schema is None:
                    raise AnonCredsObjectNotFound(
                        f"Schema not found: {schema_id}",
                        {"ledger_id": ledger_id},
                    )

                anoncreds_schema = AnonCredsSchema(
                    issuer_id=schema["id"].split(":")[0],
                    attr_names=schema["attrNames"],
                    name=schema["name"],
                    version=schema["version"],
                )
                result = GetSchemaResult(
                    schema=anoncreds_schema,
                    schema_id=schema["id"],
                    resolution_metadata={"ledger_id": ledger_id},
                    schema_metadata={"seqNo": schema["seqNo"]},
                )
            except LedgerError as err:
                raise AnonCredsResolutionError("Failed to retrieve schema") from err

        return result

    async def register_schema(
        self,
        profile: Profile,
        schema: AnonCredsSchema,
        options: Optional[dict] = None,
    ) -> SchemaResult:
        """Register a schema on the registry."""
        options = options or {}
        schema_id = self.make_schema_id(schema)

        # Assume endorser role on the network, no option for 3rd-party endorser
        ledger = profile.inject_or(BaseLedger)
        if not ledger:
            raise AnonCredsRegistrationError(NO_LEDGER_AVAILABLE_MSG)

        # Translate schema into format expected by Indy
        LOGGER.debug("Registering schema: %s", schema_id)
        indy_schema = {
            "ver": "1.0",
            "id": schema_id,
            "name": schema.name,
            "version": schema.version,
            "attrNames": schema.attr_names,
            "seqNo": None,
        }
        LOGGER.debug("schema value: %s", indy_schema)

        endorser_did = None
        create_transaction = options.get("create_transaction_for_endorser", False)

        if is_author_role(profile) or create_transaction:
            endorser_did, endorser_connection_id = await get_endorser_info(
                profile, options
            )

        write_ledger = True if endorser_did is None and not create_transaction else False

        # Get either the transaction or the seq_no or the created schema
        async with ledger:
            try:
                result = await shield(
                    ledger.send_schema_anoncreds(
                        schema_id,
                        indy_schema,
                        write_ledger=write_ledger,
                        endorser_did=endorser_did,
                    )
                )
            except LedgerObjectAlreadyExistsError as err:
                raise AnonCredsSchemaAlreadyExists(err.message, err.obj_id, schema)
            except (AnonCredsIssuerError, LedgerError) as err:
                raise AnonCredsRegistrationError("Failed to register schema") from err

        # Didn't need endorsement, so return schema result
        if write_ledger:
            return SchemaResult(
                job_id=None,
                schema_state=SchemaState(
                    state=SchemaState.STATE_FINISHED,
                    schema_id=schema_id,
                    schema=schema,
                ),
                registration_metadata={},
                schema_metadata={"seqNo": result},
            )

        # Need endorsement, so execute transaction flow
        (schema_id, schema_def) = result

        job_id = uuid4().hex
        meta_data = {"context": {"job_id": job_id, "schema_id": schema_id}}

        transaction_manager = TransactionManager(profile)
        try:
            transaction = await transaction_manager.create_record(
                messages_attach=schema_def["signed_txn"],
                connection_id=endorser_connection_id,
                meta_data=meta_data,
            )
        except StorageError:
            raise AnonCredsRegistrationError(FAILED_TO_STORE_TRANSACTION_RECORD)

        if profile.settings.get(ENDORSER_AUTO):
            try:
                (
                    transaction,
                    transaction_request,
                ) = await transaction_manager.create_request(transaction=transaction)
            except (StorageError, TransactionManagerError) as err:
                raise AnonCredsRegistrationError(
                    TRANSACTION_MANAGER_FAILED_MSG + err.roll_up
                ) from err

            responder = profile.inject(BaseResponder)
            await responder.send(
                message=transaction_request,
                connection_id=endorser_connection_id,
            )

        return SchemaResult(
            job_id=job_id,
            schema_state=SchemaState(
                state=SchemaState.STATE_WAIT,
                schema_id=schema_id,
                schema=schema,
            ),
            registration_metadata={
                "txn": transaction.serialize(),
            },
        )

    async def get_credential_definition(
        self, profile: Profile, cred_def_id: str
    ) -> GetCredDefResult:
        """Get a credential definition from the registry."""

        async with profile.session() as session:
            multitenant_mgr = session.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
            else:
                ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)

        ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
            cred_def_id,
            txn_record_type=GET_CRED_DEF,
        )
        if not ledger:
            reason = NO_LEDGER_AVAILABLE_MSG
            if not profile.settings.get_value(WALLET_TYPE):
                reason += MISSING_WALLET_TYPE_MSG
            raise AnonCredsResolutionError(reason)

        async with ledger:
            cred_def = await ledger.get_credential_definition(cred_def_id)

            if cred_def is None:
                raise AnonCredsObjectNotFound(
                    f"Credential definition not found: {cred_def_id}",
                    {"ledger_id": ledger_id},
                )

            cred_def_value = CredDefValue.deserialize(cred_def["value"])
            anoncreds_credential_definition = CredDef(
                issuer_id=cred_def["id"].split(":")[0],
                schema_id=cred_def["schemaId"],
                type=cred_def["type"],
                tag=cred_def["tag"],
                value=cred_def_value,
            )
            anoncreds_registry_get_credential_definition = GetCredDefResult(
                credential_definition=anoncreds_credential_definition,
                credential_definition_id=cred_def["id"],
                resolution_metadata={},
                credential_definition_metadata={},
            )
        return anoncreds_registry_get_credential_definition

    async def register_credential_definition(
        self,
        profile: Profile,
        schema: GetSchemaResult,
        credential_definition: CredDef,
        options: Optional[dict] = None,
    ) -> CredDefResult:
        """Register a credential definition on the registry."""
        options = options or {}
        cred_def_id = self.make_cred_def_id(schema, credential_definition)

        ledger = profile.inject_or(BaseLedger)
        if not ledger:
            raise AnonCredsRegistrationError(NO_LEDGER_AVAILABLE_MSG)

        # Check if in wallet but not on ledger
        issuer = AnonCredsIssuer(profile)
        if await issuer.credential_definition_in_wallet(cred_def_id):
            try:
                await self.get_credential_definition(profile, cred_def_id)
            except AnonCredsObjectNotFound as err:
                raise AnonCredsRegistrationError(
                    f"Credential definition with id {cred_def_id} already "
                    "exists in wallet but not on the ledger"
                ) from err

        # Translate anoncreds object to indy object
        LOGGER.debug("Registering credential definition: %s", cred_def_id)
        indy_cred_def = {
            "id": cred_def_id,
            "schemaId": str(schema.schema_metadata["seqNo"]),
            "tag": credential_definition.tag,
            "type": credential_definition.type,
            "value": credential_definition.value.serialize(),
            "ver": "1.0",
        }
        LOGGER.debug("Cred def value: %s", indy_cred_def)

        endorser_did = None
        create_transaction = options.get("create_transaction_for_endorser", False)

        if is_author_role(profile) or create_transaction:
            endorser_did, endorser_connection_id = await get_endorser_info(
                profile, options
            )

        write_ledger = True if endorser_did is None and not create_transaction else False

        async with ledger:
            try:
                result = await shield(
                    ledger.send_credential_definition_anoncreds(
                        credential_definition.schema_id,
                        cred_def_id,
                        indy_cred_def,
                        write_ledger=write_ledger,
                        endorser_did=endorser_did,
                    )
                )
            except LedgerObjectAlreadyExistsError as err:
                if await issuer.credential_definition_in_wallet(cred_def_id):
                    raise AnonCredsObjectAlreadyExists(
                        f"Credential definition with id {cred_def_id} "
                        "already exists in wallet and on ledger.",
                        cred_def_id,
                    ) from err
                else:
                    raise AnonCredsObjectAlreadyExists(
                        f"Credential definition {cred_def_id} is on "
                        f"ledger but not in wallet {profile.name}",
                        cred_def_id,
                    ) from err

        # Didn't need endorsement
        if write_ledger:
            return CredDefResult(
                job_id=None,
                credential_definition_state=CredDefState(
                    state=CredDefState.STATE_FINISHED,
                    credential_definition_id=cred_def_id,
                    credential_definition=credential_definition,
                ),
                registration_metadata={},
                credential_definition_metadata={"seqNo": result},
            )

        # Need endorsement, so execute transaction flow
        job_id = uuid4().hex

        meta_data = {
            "context": {
                "job_id": job_id,
                "cred_def_id": cred_def_id,
                "options": options,
            },
        }

        (cred_def_id, cred_def) = result
        transaction_manager = TransactionManager(profile)

        try:
            transaction = await transaction_manager.create_record(
                messages_attach=cred_def["signed_txn"],
                connection_id=endorser_connection_id,
                meta_data=meta_data,
            )
        except StorageError:
            raise AnonCredsRegistrationError(FAILED_TO_STORE_TRANSACTION_RECORD)

        if profile.settings.get(ENDORSER_AUTO):
            try:
                (
                    transaction,
                    transaction_request,
                ) = await transaction_manager.create_request(transaction=transaction)
            except (StorageError, TransactionManagerError) as err:
                raise AnonCredsRegistrationError(
                    TRANSACTION_MANAGER_FAILED_MSG + err.roll_up
                ) from err

            responder = profile.inject(BaseResponder)
            await responder.send(
                message=transaction_request,
                connection_id=endorser_connection_id,
            )

        return CredDefResult(
            job_id=job_id,
            credential_definition_state=CredDefState(
                state=CredDefState.STATE_WAIT,
                credential_definition_id=cred_def_id,
                credential_definition=credential_definition,
            ),
            registration_metadata={
                "txn": transaction.serialize(),
            },
            credential_definition_metadata={},
        )

    async def get_revocation_registry_definition(
        self, profile: Profile, rev_reg_def_id: str
    ) -> GetRevRegDefResult:
        """Get a revocation registry definition from the registry."""
        async with profile.session() as session:
            multitenant_mgr = session.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
            else:
                ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)

        ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
            rev_reg_def_id,
            txn_record_type=GET_CRED_DEF,
        )
        if not ledger:
            reason = NO_LEDGER_AVAILABLE_MSG
            if not profile.settings.get_value(WALLET_TYPE):
                reason += MISSING_WALLET_TYPE_MSG
            raise AnonCredsResolutionError(reason)

        async with ledger:
            rev_reg_def = await ledger.get_revoc_reg_def(rev_reg_def_id)

            if rev_reg_def is None:
                raise AnonCredsObjectNotFound(
                    f"Revocation registry definition not found: {rev_reg_def_id}",
                    {"ledger_id": ledger_id},
                )

            LOGGER.debug("Retrieved revocation registry definition: %s", rev_reg_def)
            rev_reg_def_value = RevRegDefValue.deserialize(rev_reg_def["value"])
            anoncreds_rev_reg_def = RevRegDef(
                issuer_id=rev_reg_def["id"].split(":")[0],
                cred_def_id=rev_reg_def["credDefId"],
                type=rev_reg_def["revocDefType"],
                value=rev_reg_def_value,
                tag=rev_reg_def["tag"],
            )
            result = GetRevRegDefResult(
                revocation_registry=anoncreds_rev_reg_def,
                revocation_registry_id=rev_reg_def["id"],
                resolution_metadata={},
                revocation_registry_metadata={},
            )

        return result

    async def register_revocation_registry_definition(
        self,
        profile: Profile,
        revocation_registry_definition: RevRegDef,
        options: Optional[dict] = None,
    ) -> RevRegDefResult:
        """Register a revocation registry definition on the registry."""
        options = options or {}
        rev_reg_def_id = self.make_rev_reg_def_id(revocation_registry_definition)

        ledger = profile.inject(BaseLedger)
        if not ledger:
            raise AnonCredsRegistrationError(NO_LEDGER_AVAILABLE_MSG)

        # Translate anoncreds object to indy object
        indy_rev_reg_def = {
            "ver": "1.0",
            "id": rev_reg_def_id,
            "revocDefType": revocation_registry_definition.type,
            "credDefId": revocation_registry_definition.cred_def_id,
            "tag": revocation_registry_definition.tag,
            "value": {
                "issuanceType": "ISSUANCE_BY_DEFAULT",
                "maxCredNum": revocation_registry_definition.value.max_cred_num,
                "publicKeys": revocation_registry_definition.value.public_keys,
                "tailsHash": revocation_registry_definition.value.tails_hash,
                "tailsLocation": revocation_registry_definition.value.tails_location,
            },
        }

        endorser_did = None
        create_transaction = options.get(
            "create_transaction_for_endorser", False
        ) or profile.settings.get("endorser.auto_create_rev_reg", False)

        if is_author_role(profile) or create_transaction:
            endorser_did, endorser_connection_id = await get_endorser_info(
                profile, options
            )

        write_ledger = True if endorser_did is None and not create_transaction else False

        try:
            async with ledger:
                result = await ledger.send_revoc_reg_def(
                    indy_rev_reg_def,
                    revocation_registry_definition.issuer_id,
                    write_ledger=write_ledger,
                    endorser_did=endorser_did,
                )
        except LedgerError as err:
            LOGGER.error(
                f"Error registering revocation registry definition {rev_reg_def_id}: {err.roll_up}"  # noqa: E501
            )

        # Didn't need endorsement
        if write_ledger:
            return RevRegDefResult(
                job_id=None,
                revocation_registry_definition_state=RevRegDefState(
                    state=RevRegDefState.STATE_FINISHED,
                    revocation_registry_definition_id=rev_reg_def_id,
                    revocation_registry_definition=revocation_registry_definition,
                ),
                registration_metadata={},
                revocation_registry_definition_metadata={"seqNo": result},
            )

        # Need endorsement, so execute transaction flow
        (rev_reg_def_id, reg_rev_def) = result

        job_id = uuid4().hex
        meta_data = {
            "context": {
                "job_id": job_id,
                "rev_reg_def_id": rev_reg_def_id,
                "options": options,
            }
        }

        transaction_manager = TransactionManager(profile)
        try:
            transaction = await transaction_manager.create_record(
                messages_attach=reg_rev_def["signed_txn"],
                connection_id=endorser_connection_id,
                meta_data=meta_data,
            )
        except StorageError:
            raise AnonCredsRegistrationError(FAILED_TO_STORE_TRANSACTION_RECORD)

        if profile.settings.get(ENDORSER_AUTO):
            try:
                (
                    transaction,
                    transaction_request,
                ) = await transaction_manager.create_request(transaction=transaction)
            except (StorageError, TransactionManagerError) as err:
                raise AnonCredsRegistrationError(
                    TRANSACTION_MANAGER_FAILED_MSG + err.roll_up
                ) from err

            responder = profile.inject(BaseResponder)
            await responder.send(
                message=transaction_request,
                connection_id=endorser_connection_id,
            )

        return RevRegDefResult(
            job_id=job_id,
            revocation_registry_definition_state=RevRegDefState(
                state=RevRegDefState.STATE_WAIT,
                revocation_registry_definition_id=rev_reg_def_id,
                revocation_registry_definition=revocation_registry_definition,
            ),
            registration_metadata={
                "txn": transaction.serialize(),
            },
            revocation_registry_definition_metadata={},
        )

    async def _get_or_fetch_rev_reg_def_max_cred_num(
        self, profile: Profile, ledger: BaseLedger, rev_reg_def_id: str
    ) -> int:
        """Retrieve max cred num for a rev reg def.

        The value is retrieved from cache or from the ledger if necessary.
        The issuer could retrieve this value from the wallet but this info
        must also be known to the holder.
        """
        cache = profile.inject(BaseCache)
        cache_key = f"anoncreds::legacy_indy::rev_reg_max_cred_num::{rev_reg_def_id}"

        if cache:
            max_cred_num = await cache.get(cache_key)
            if max_cred_num:
                return max_cred_num

        rev_reg_def = await ledger.get_revoc_reg_def(rev_reg_def_id)
        max_cred_num = rev_reg_def["value"]["maxCredNum"]

        if cache:
            await cache.set(cache_key, max_cred_num)

        return max_cred_num

    def _indexes_to_bit_array(self, indexes: List[int], size: int) -> List[int]:
        """Turn a sequence of indexes into a full state bit array."""
        return [1 if index in indexes else 0 for index in range(0, size + 1)]

    async def _get_ledger(self, profile: Profile, rev_reg_def_id: str):
        async with profile.session() as session:
            multitenant_mgr = session.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
            else:
                ledger_exec_inst = session.inject(IndyLedgerRequestsExecutor)

        ledger_id, ledger = await ledger_exec_inst.get_ledger_for_identifier(
            rev_reg_def_id,
            txn_record_type=GET_CRED_DEF,
        )
        if not ledger:
            reason = NO_LEDGER_AVAILABLE_MSG
            if not profile.settings.get_value(WALLET_TYPE):
                reason += MISSING_WALLET_TYPE_MSG
            raise AnonCredsResolutionError(reason)

        return ledger_id, ledger

    async def get_revocation_registry_delta(
        self,
        profile: Profile,
        rev_reg_def_id: str,
        timestamp_from: Optional[int] = 0,
        timestamp_to: Optional[int] = None,
    ) -> Tuple[dict, int]:
        """Fetch the revocation registry delta."""
        ledger_id, ledger = await self._get_ledger(profile, rev_reg_def_id)

        async with ledger:
            delta, timestamp = await ledger.get_revoc_reg_delta(
                rev_reg_def_id, timestamp_from=timestamp_from, timestamp_to=timestamp_to
            )

            if delta is None:
                raise AnonCredsObjectNotFound(
                    f"Revocation list not found for rev reg def: {rev_reg_def_id}",
                    {"ledger_id": ledger_id},
                )
        LOGGER.debug("Retrieved delta: %s", delta)
        return delta, timestamp

    async def get_revocation_list(
        self,
        profile: Profile,
        rev_reg_def_id: str,
        timestamp_from: Optional[int] = 0,
        timestamp_to: Optional[int] = None,
    ) -> GetRevListResult:
        """Get the revocation registry list."""
        _, ledger = await self._get_ledger(profile, rev_reg_def_id)

        delta, timestamp = await self.get_revocation_registry_delta(
            profile, rev_reg_def_id, timestamp_from, timestamp_to
        )

        async with ledger:
            max_cred_num = await self._get_or_fetch_rev_reg_def_max_cred_num(
                profile, ledger, rev_reg_def_id
            )
            revocation_list_from_indexes = self._indexes_to_bit_array(
                delta["value"]["revoked"], max_cred_num
            )
            LOGGER.debug(
                "Index list to full state bit array: %s", revocation_list_from_indexes
            )
            rev_list = RevList(
                issuer_id=rev_reg_def_id.split(":")[0],
                rev_reg_def_id=rev_reg_def_id,
                revocation_list=revocation_list_from_indexes,
                current_accumulator=delta["value"]["accum"],
                timestamp=timestamp,
            )
            result = GetRevListResult(
                revocation_list=rev_list,
                resolution_metadata={},
                revocation_registry_metadata={},
            )

        return result

    async def _revoc_reg_entry_with_fix(
        self,
        profile: Profile,
        rev_list: RevList,
        rev_reg_def_type: str,
        entry: dict,
        write_ledger: bool,
        endorser_did: Optional[str] = None,
    ) -> dict:
        """Send a revocation registry entry to the ledger with fixes if needed."""
        multitenant_mgr = profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
        else:
            ledger_exec_inst = profile.inject(IndyLedgerRequestsExecutor)
        _, ledger = await ledger_exec_inst.get_ledger_for_identifier(
            rev_list.rev_reg_def_id,
            txn_record_type=GET_REVOC_REG_ENTRY,
        )

        try:
            async with ledger:
                rev_entry_res = await ledger.send_revoc_reg_entry(
                    rev_list.rev_reg_def_id,
                    rev_reg_def_type,
                    entry,
                    rev_list.issuer_id,
                    write_ledger=write_ledger,
                    endorser_did=endorser_did,
                )
        except LedgerTransactionError as err:
            if "InvalidClientRequest" in err.roll_up:
                # ... if the ledger write fails (with "InvalidClientRequest")
                # e.g. acapy_agent.ledger.error.LedgerTransactionError:
                #   Ledger rejected transaction request: client request invalid:
                #   InvalidClientRequest(...)
                # In this scenario we try to post a correction
                LOGGER.warning("Retry ledger update/fix due to error")
                LOGGER.warning(err)
                (_, _, rev_entry_res) = await self.fix_ledger_entry(
                    profile,
                    rev_list,
                    True,
                    ledger.pool.genesis_txns,
                    write_ledger,
                    endorser_did,
                )
                LOGGER.warning("Ledger update/fix applied")
            elif "InvalidClientTaaAcceptanceError" in err.roll_up:
                # if no write access (with "InvalidClientTaaAcceptanceError")
                # e.g. acapy_agent.ledger.error.LedgerTransactionError:
                #   Ledger rejected transaction request: client request invalid:
                #   InvalidClientTaaAcceptanceError(...)
                LOGGER.exception("Ledger update failed due to TAA issue")
                raise AnonCredsRegistrationError(
                    "Ledger update failed due to TAA Issue"
                ) from err
            else:
                LOGGER.exception("Ledger update failed due to unknown issue")
                raise AnonCredsRegistrationError(
                    "Ledger update failed due to unknown issue"
                ) from err

        return rev_entry_res

    async def register_revocation_list(
        self,
        profile: Profile,
        rev_reg_def: RevRegDef,
        rev_list: RevList,
        options: Optional[dict] = None,
    ) -> RevListResult:
        """Register a revocation list on the registry."""
        options = options or {}
        rev_reg_entry = {"ver": "1.0", "value": {"accum": rev_list.current_accumulator}}

        endorser_did = None
        create_transaction = options.get(
            "create_transaction_for_endorser", False
        ) or profile.settings.get("endorser.auto_create_rev_reg", False)

        if is_author_role(profile) or create_transaction:
            endorser_did, endorser_connection_id = await get_endorser_info(
                profile, options
            )

        write_ledger = True if endorser_did is None and not create_transaction else False

        result = await self._revoc_reg_entry_with_fix(
            profile,
            rev_list,
            rev_reg_def.type,
            rev_reg_entry,
            write_ledger,
            endorser_did,
        )

        if write_ledger:
            return RevListResult(
                job_id=None,
                revocation_list_state=RevListState(
                    state=RevListState.STATE_FINISHED,
                    revocation_list=rev_list,
                ),
                registration_metadata={},
                revocation_list_metadata={"seqNo": result},
            )

        (rev_reg_def_id, requested_txn) = result

        job_id = uuid4().hex
        meta_data = {
            "context": {
                "job_id": job_id,
                "rev_reg_def_id": rev_reg_def_id,
                "options": {
                    "endorser_connection_id": endorser_connection_id,
                    "create_transaction_for_endorser": create_transaction,
                },
            }
        }

        transaction_manager = TransactionManager(profile)
        try:
            transaction = await transaction_manager.create_record(
                messages_attach=requested_txn["signed_txn"],
                connection_id=endorser_connection_id,
                meta_data=meta_data,
            )
        except StorageError:
            raise AnonCredsRegistrationError(FAILED_TO_STORE_TRANSACTION_RECORD)

        if profile.settings.get(ENDORSER_AUTO):
            try:
                (
                    transaction,
                    transaction_request,
                ) = await transaction_manager.create_request(transaction=transaction)
            except (StorageError, TransactionManagerError) as err:
                raise AnonCredsRegistrationError(
                    TRANSACTION_MANAGER_FAILED_MSG + err.roll_up
                ) from err

            responder = profile.inject(BaseResponder)
            await responder.send(
                message=transaction_request,
                connection_id=endorser_connection_id,
            )

        return RevListResult(
            job_id=job_id,
            revocation_list_state=RevListState(
                state=RevListState.STATE_WAIT,
                revocation_list=rev_list,
            ),
            registration_metadata={
                "txn": transaction.serialize(),
            },
            revocation_list_metadata={},
        )

    async def update_revocation_list(
        self,
        profile: Profile,
        rev_reg_def: RevRegDef,
        prev_list: RevList,
        curr_list: RevList,
        revoked: Sequence[int],
        options: Optional[dict] = None,
    ) -> RevListResult:
        """Update a revocation list."""
        options = options or {}
        newly_revoked_indices = list(revoked)
        rev_reg_entry = {
            "ver": "1.0",
            "value": {
                "accum": curr_list.current_accumulator,
                "prevAccum": prev_list.current_accumulator,
                "revoked": newly_revoked_indices,
            },
        }

        endorser_did = None
        create_transaction = options.get("create_transaction_for_endorser", False)

        if is_author_role(profile) or create_transaction:
            endorser_did, endorser_connection_id = await get_endorser_info(
                profile, options
            )

        write_ledger = True if endorser_did is None and not create_transaction else False

        result = await self._revoc_reg_entry_with_fix(
            profile,
            curr_list,
            rev_reg_def.type,
            rev_reg_entry,
            write_ledger,
            endorser_did,
        )

        if write_ledger:
            event_bus = profile.inject(EventBus)
            await event_bus.notify(
                profile,
                RevListFinishedEvent.with_payload(
                    curr_list.rev_reg_def_id, newly_revoked_indices
                ),
            )
            return RevListResult(
                job_id=None,
                revocation_list_state=RevListState(
                    state=RevListState.STATE_FINISHED,
                    revocation_list=curr_list,
                ),
                registration_metadata={},
                revocation_list_metadata={"seqNo": result},
            )

        (rev_reg_def_id, requested_txn) = result

        job_id = uuid4().hex
        meta_data = {
            "context": {
                "job_id": job_id,
                "rev_reg_def_id": rev_reg_def_id,
                "rev_list": curr_list.serialize(),
                "options": {
                    "endorser_connection_id": endorser_connection_id,
                    "create_transaction_for_endorser": create_transaction,
                },
            }
        }

        transaction_manager = TransactionManager(profile)
        try:
            transaction = await transaction_manager.create_record(
                messages_attach=requested_txn["signed_txn"],
                connection_id=endorser_connection_id,
                meta_data=meta_data,
            )
        except StorageError:
            raise AnonCredsRegistrationError(FAILED_TO_STORE_TRANSACTION_RECORD)

        if profile.settings.get(ENDORSER_AUTO):
            try:
                (
                    transaction,
                    transaction_request,
                ) = await transaction_manager.create_request(transaction=transaction)
            except (StorageError, TransactionManagerError) as err:
                raise AnonCredsRegistrationError(
                    TRANSACTION_MANAGER_FAILED_MSG + err.roll_up
                ) from err

            responder = profile.inject(BaseResponder)
            await responder.send(
                message=transaction_request,
                connection_id=endorser_connection_id,
            )

        return RevListResult(
            job_id=job_id,
            revocation_list_state=RevListState(
                state=RevListState.STATE_WAIT,
                revocation_list=curr_list,
            ),
            registration_metadata={
                "txn": transaction.serialize(),
            },
            revocation_list_metadata={},
        )

    async def fix_ledger_entry(
        self,
        profile: Profile,
        rev_list: RevList,
        apply_ledger_update: bool,
        genesis_transactions: str,
        write_ledger: bool = True,
        endorser_did: Optional[str] = None,
    ) -> Tuple[dict, dict, dict]:
        """Fix the ledger entry to match wallet-recorded credentials."""

        def _wallet_accumalator_matches_ledger_list(
            rev_list: RevList, rev_reg_delta: dict
        ) -> bool:
            return (
                rev_reg_delta.get("value")
                and rev_list.current_accumulator == rev_reg_delta["value"]["accum"]
            )

        applied_txn = {}
        recovery_txn = {}

        LOGGER.debug("Fixing ledger entry for revocation list...")

        multitenant_mgr = profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(profile)
        else:
            ledger_exec_inst = profile.inject(IndyLedgerRequestsExecutor)
        _, ledger = await ledger_exec_inst.get_ledger_for_identifier(
            rev_list.rev_reg_def_id,
            txn_record_type=GET_REVOC_REG_DELTA,
        )

        async with ledger:
            (rev_reg_delta, _) = await ledger.get_revoc_reg_delta(rev_list.rev_reg_def_id)

        async with profile.session() as session:
            LOGGER.debug(f"revocation_list = {rev_list.revocation_list}")
            LOGGER.debug(f"rev_reg_delta = {rev_reg_delta.get('value')}")

            rev_list = await self._sync_wallet_rev_list_with_issuer_cred_rev_records(
                session, rev_list
            )

            if not _wallet_accumalator_matches_ledger_list(rev_list, rev_reg_delta):
                recovery_txn = await generate_ledger_rrrecovery_txn(
                    genesis_transactions, rev_list
                )

                if apply_ledger_update and recovery_txn:
                    ledger = session.inject_or(BaseLedger)
                    if not ledger:
                        reason = NO_LEDGER_AVAILABLE_MSG
                        if not session.context.settings.get_value(WALLET_TYPE):
                            reason += MISSING_WALLET_TYPE_MSG
                        raise LedgerError(reason=reason)

                    async with ledger:
                        applied_txn = await ledger.send_revoc_reg_entry(
                            rev_list.rev_reg_def_id,
                            "CL_ACCUM",
                            recovery_txn,
                            rev_list.issuer_id,
                            write_ledger,
                            endorser_did,
                        )

        return (rev_reg_delta, recovery_txn, applied_txn)

    async def _sync_wallet_rev_list_with_issuer_cred_rev_records(
        self, session: ProfileSession, rev_list: RevList
    ) -> RevList:
        """Sync the wallet revocation list with the issuer cred rev records."""

        async def _revoked_issuer_cred_rev_record_ids() -> List[int]:
            cred_rev_records = await IssuerCredRevRecord.query_by_ids(
                session, rev_reg_id=rev_list.rev_reg_def_id
            )
            return [
                int(rec.cred_rev_id) for rec in cred_rev_records if rec.state == "revoked"
            ]

        def _revocation_list_to_array_of_indexes(
            revocation_list: List[int],
        ) -> List[int]:
            return [index for index, value in enumerate(revocation_list) if value == 1]

        revoked = await _revoked_issuer_cred_rev_record_ids()
        if revoked == _revocation_list_to_array_of_indexes(rev_list.revocation_list):
            return rev_list

        # The revocation list is out of sync with the issuer cred rev records
        # Recreate the revocation list with the issuer cred rev records
        revoc_reg_def_entry = await session.handle.fetch(
            CATEGORY_REV_REG_DEF, rev_list.rev_reg_def_id
        )
        cred_def_entry = await session.handle.fetch(
            CATEGORY_CRED_DEF,
            RevRegDef.deserialize(revoc_reg_def_entry.value_json).cred_def_id,
        )
        revoc_reg_def_private_entry = await session.handle.fetch(
            CATEGORY_REV_REG_DEF_PRIVATE, rev_list.rev_reg_def_id
        )
        updated_list = await asyncio.get_event_loop().run_in_executor(
            None,
            lambda: rev_list.to_native().update(
                cred_def=CredentialDefinition.load(cred_def_entry.value_json),
                rev_reg_def=RevocationRegistryDefinition.load(
                    revoc_reg_def_entry.value_json
                ),
                rev_reg_def_private=RevocationRegistryDefinitionPrivate.load(
                    revoc_reg_def_private_entry.raw_value
                ),
                issued=None,
                revoked=revoked,
                timestamp=None,
            ),
        )
        rev_list_entry_update = await session.handle.fetch(
            CATEGORY_REV_LIST, rev_list.rev_reg_def_id, for_update=True
        )
        tags = rev_list_entry_update.tags
        rev_list_entry_update = rev_list_entry_update.value_json
        rev_list_entry_update["rev_list"] = updated_list.to_dict()

        await session.handle.replace(
            CATEGORY_REV_LIST,
            rev_list.rev_reg_def_id,
            value_json=rev_list_entry_update,
            tags=tags,
        )
        return RevList.deserialize(updated_list.to_json())

    async def txn_submit(
        self,
        ledger: BaseLedger,
        ledger_transaction: str,
        sign: Optional[bool] = None,
        taa_accept: Optional[bool] = None,
        sign_did: DIDInfo = sentinel,
        write_ledger: bool = True,
    ) -> str:
        """Submit a transaction to the ledger."""

        try:
            async with ledger:
                return await shield(
                    ledger.txn_submit(
                        ledger_transaction,
                        sign=sign,
                        taa_accept=taa_accept,
                        sign_did=sign_did,
                        write_ledger=write_ledger,
                    )
                )
        except LedgerError as err:
            raise AnonCredsRegistrationError(err.roll_up) from err

    async def get_schema_info_by_id(
        self, profile: Profile, schema_id: str
    ) -> AnonCredsSchemaInfo:
        """Get schema info by schema id."""
        schema_id_parts = re.match(r"^(\w+):2:([^:]+):([^:]+)$", schema_id)
        return AnonCredsSchemaInfo(
            issuer_id=schema_id_parts.group(1),
            name=schema_id_parts.group(2),
            version=schema_id_parts.group(3),
        )



"""Presentation ack message handler."""

from .....core.oob_processor import OobMessageProcessor
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....utils.tracing import get_timer, trace_event
from ..manager import PresentationManager
from ..messages.presentation_ack import PresentationAck


class PresentationAckHandler(BaseHandler):
    """Message handler class for presentation acks."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for presentation acks.

        Args:
            context: request context
            responder: responder callback
        """
        r_time = get_timer()

        self._logger.debug("PresentationAckHandler called with context %s", context)
        assert isinstance(context.message, PresentationAck)
        self._logger.debug(
            "Received presentation ack message: %s",
            context.message.serialize(as_string=True),
        )

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException("Connection used for presentation ack not ready")

        # Find associated oob record
        oob_processor = context.inject(OobMessageProcessor)
        oob_record = await oob_processor.find_oob_record_for_inbound_message(context)

        # Either connection or oob context must be present
        if not context.connection_record and not oob_record:
            raise HandlerException(
                "No connection or associated connectionless exchange found for"
                " presentation ack"
            )

        presentation_manager = PresentationManager(context.profile)
        await presentation_manager.receive_presentation_ack(
            context.message, context.connection_record
        )

        trace_event(
            context.settings,
            context.message,
            outcome="PresentationAckHandler.handle.END",
            perf_counter=r_time,
        )



from unittest import IsolatedAsyncioTestCase, TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import ENDORSED_TRANSACTION_RESPONSE
from ..endorsed_transaction_response import EndorsedTransactionResponse


class TestConfig:
    test_transaction_id = "3fa85f64-5717-4562-b3fc-2c963f66afa6"
    test_thread_id = "3fa85f64-5717-4562-b3fc-2c963f66afa6"
    test_signature_response = {
        "message_id": "95f890bf-329f-4450-9d9d-89082077acf9",
        "context": "did:sov",
        "method": "add-signature",
        "signer_goal_code": "transaction.endorse",
        "signature_type": "<requested signature type>",
        "signature": {
            "4Sr8NmbMuq6SHLrh7Bv5sN": "2svZP3omPic2Y9xQMh2AVhTd6Hhbrt6Yxc6d9cbwNGLu"
        },
    }
    test_state = "endorsed"
    test_endorser_did = "V4SGRU86Z58d6TV7PBUe6f"


class TestEndorsedTransactionResponse(TestCase, TestConfig):
    def setUp(self):
        self.endorsed_transaction_response = EndorsedTransactionResponse(
            transaction_id=self.test_transaction_id,
            thread_id=self.test_thread_id,
            signature_response=self.test_signature_response,
            state=self.test_state,
            endorser_did=self.test_endorser_did,
        )

    def test_init(self):
        """Test initialization."""
        assert (
            self.endorsed_transaction_response.transaction_id == self.test_transaction_id
        )
        assert self.endorsed_transaction_response.thread_id == self.test_thread_id
        assert (
            self.endorsed_transaction_response.signature_response
            == self.test_signature_response
        )
        assert self.endorsed_transaction_response.state == self.test_state
        assert self.endorsed_transaction_response.endorser_did == self.test_endorser_did

    def test_type(self):
        """Test type."""
        assert self.endorsed_transaction_response._type == DIDCommPrefix.qualify_current(
            ENDORSED_TRANSACTION_RESPONSE
        )

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "endorsed_transaction_response.EndorsedTransactionResponseSchema.load"
    )
    def test_deserialize(self, mock_endorsed_transaction_response_schema_load):
        """
        Test deserialization.
        """
        obj = self.endorsed_transaction_response

        endorsed_transaction_response = EndorsedTransactionResponse.deserialize(obj)
        mock_endorsed_transaction_response_schema_load.assert_called_once_with(obj)

        assert (
            endorsed_transaction_response
            is mock_endorsed_transaction_response_schema_load.return_value
        )

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "endorsed_transaction_response.EndorsedTransactionResponseSchema.dump"
    )
    def test_serialize(self, mock_endorsed_transaction_response_schema_dump):
        """
        Test serialization.
        """
        endorsed_transaction_response_dict = (
            self.endorsed_transaction_response.serialize()
        )
        mock_endorsed_transaction_response_schema_dump.assert_called_once_with(
            self.endorsed_transaction_response
        )

        assert (
            endorsed_transaction_response_dict
            is mock_endorsed_transaction_response_schema_dump.return_value
        )


class TestEndorsedTransactionResponseSchema(IsolatedAsyncioTestCase, TestConfig):
    """Test endorsed transaction response schema."""

    async def test_make_model(self):
        endorsed_transaction_response = EndorsedTransactionResponse(
            transaction_id=self.test_transaction_id,
            thread_id=self.test_thread_id,
            signature_response=self.test_signature_response,
            state=self.test_state,
            endorser_did=self.test_endorser_did,
        )
        data = endorsed_transaction_response.serialize()
        model_instance = EndorsedTransactionResponse.deserialize(data)
        assert type(model_instance) is type(endorsed_transaction_response)



"""DID INDY routes."""

from http import HTTPStatus

from aiohttp import web
from aiohttp_apispec import docs, request_schema, response_schema
from marshmallow import fields

from ...admin.decorators.auth import tenant_authentication
from ...admin.request_context import AdminRequestContext
from ...did.indy.indy_manager import DidIndyManager
from ...messaging.models.openapi import OpenAPISchema
from ...wallet.error import WalletError


class CreateDidIndyRequestSchema(OpenAPISchema):
    """Parameters and validators for create DID Indy endpoint."""

    options = fields.Dict(
        required=False,
        metadata={
            "description": (
                "Additional configuration options. "
                "Supported options: did, seed, key_type. "
                "Default key_type is ed25519."
            ),
            "example": {
                "did": "did:indy:WRfXPg8dantKVubE3HX8pw",
                "seed": "000000000000000000000000Trustee1",
                "key_type": "ed25519",
            },
        },
    )
    features = fields.Dict(
        required=False,
        metadata={
            "description": "Additional features to enable for the did.",
            "example": "{}",
        },
    )


class CreateDidIndyResponseSchema(OpenAPISchema):
    """Response schema for create DID Indy endpoint."""

    did = fields.Str(
        metadata={
            "description": "DID created",
            "example": "did:indy:DFZgMggBEXcZFVQ2ZBTwdr",
        }
    )
    verkey = fields.Str(
        metadata={
            "description": "Verification key",
            "example": "BnSWTUQmdYCewSGFrRUhT6LmKdcCcSzRGqWXMPnEP168",
        }
    )


@docs(tags=["did"], summary="Create a did:indy")
@request_schema(CreateDidIndyRequestSchema())
@response_schema(CreateDidIndyResponseSchema, HTTPStatus.OK)
@tenant_authentication
async def create_indy_did(request: web.BaseRequest):
    """Create a INDY DID."""
    context: AdminRequestContext = request["context"]
    body = await request.json()
    options = body.get("options", {})
    try:
        result = await DidIndyManager(context.profile).register(options)
        return web.json_response(result)
    except WalletError as e:
        raise web.HTTPBadRequest(reason=str(e))


async def register(app: web.Application):
    """Register routes."""
    app.add_routes([web.post("/did/indy/create", create_indy_did)])


def post_process_routes(app: web.Application):
    """Amend swagger API."""
    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "did",
            "description": "Endpoints for managing dids",
            "externalDocs": {
                "description": "Specification",
                "url": "https://www.w3.org/TR/did-core/",
            },
        }
    )



"""Credential proposal message handler."""

from .....anoncreds.issuer import AnonCredsIssuerError
from .....indy.issuer import IndyIssuerError
from .....ledger.error import LedgerError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError
from .....utils.tracing import get_timer, trace_event
from .. import problem_report_for_record
from ..manager import V20CredManager, V20CredManagerError
from ..messages.cred_problem_report import ProblemReportReason
from ..messages.cred_proposal import V20CredProposal


class V20CredProposalHandler(BaseHandler):
    """Message handler class for credential proposals."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for credential proposals.

        Args:
            context: proposal context
            responder: responder callback

        """
        r_time = get_timer()

        self._logger.debug("V20CredProposalHandler called with context %s", context)
        assert isinstance(context.message, V20CredProposal)
        self._logger.debug(
            "Received v2.0 credential proposal message: %s",
            context.message.serialize(as_string=True),
        )

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException("Connection used for credential proposal not ready")
        elif not context.connection_record:
            raise HandlerException("Connectionless not supported for credential proposal")

        profile = context.profile
        cred_manager = V20CredManager(profile)
        cred_ex_record = await cred_manager.receive_proposal(
            context.message, context.connection_record.connection_id
        )  # mgr only finds, saves record: on exception, saving state null is hopeless

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="CredentialProposalHandler.handle.END",
            perf_counter=r_time,
        )

        # If auto_offer is enabled, respond immediately with offer
        if cred_ex_record and cred_ex_record.auto_offer:
            cred_offer_message = None
            try:
                (cred_ex_record, cred_offer_message) = await cred_manager.create_offer(
                    cred_ex_record,
                    counter_proposal=None,
                    comment=context.message.comment,
                )
                await responder.send_reply(cred_offer_message)
            except (
                BaseModelError,
                AnonCredsIssuerError,
                IndyIssuerError,
                LedgerError,
                StorageError,
                V20CredManagerError,
            ) as err:
                self._logger.exception("Error responding to credential proposal")
                async with profile.session() as session:
                    await cred_ex_record.save_error_state(
                        session,
                        reason=err.roll_up,  # us: be specific
                    )
                await responder.send_reply(
                    problem_report_for_record(
                        cred_ex_record,
                        ProblemReportReason.ISSUANCE_ABANDONED.value,  # them: vague
                    )
                )

            trace_event(
                context.settings,
                cred_offer_message,
                outcome="V20CredProposalHandler.handle.OFFER",
                perf_counter=r_time,
            )



"""Wallet admin routes."""

import asyncio
import json
import logging
from typing import List, Optional, Tuple, Union

from aiohttp import web
from aiohttp_apispec import docs, querystring_schema, request_schema, response_schema
from marshmallow import fields, validate

from ..admin.decorators.auth import tenant_authentication
from ..admin.request_context import AdminRequestContext
from ..config.injection_context import InjectionContext
from ..connections.base_manager import BaseConnectionManager
from ..connections.models.conn_record import ConnRecord
from ..core.event_bus import Event, EventBus
from ..core.profile import Profile
from ..ledger.base import BaseLedger
from ..ledger.endpoint_type import EndpointType
from ..ledger.error import LedgerConfigError, LedgerError
from ..messaging.jsonld.error import BadJWSHeaderError, InvalidVerificationMethod
from ..messaging.models.base import BaseModelError
from ..messaging.models.openapi import OpenAPISchema
from ..messaging.responder import BaseResponder
from ..messaging.valid import (
    DID_POSTURE_EXAMPLE,
    DID_POSTURE_VALIDATE,
    ENDPOINT_EXAMPLE,
    ENDPOINT_TYPE_EXAMPLE,
    ENDPOINT_TYPE_VALIDATE,
    ENDPOINT_VALIDATE,
    GENERIC_DID_EXAMPLE,
    GENERIC_DID_VALIDATE,
    INDY_DID_EXAMPLE,
    INDY_DID_VALIDATE,
    JWT_EXAMPLE,
    JWT_VALIDATE,
    NON_SD_LIST_EXAMPLE,
    NON_SD_LIST_VALIDATE,
    RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
    RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
    SD_JWT_EXAMPLE,
    SD_JWT_VALIDATE,
    UUID4_EXAMPLE,
    UUID4_VALIDATE,
    IndyDID,
    StrOrDictField,
    Uri,
)
from ..protocols.coordinate_mediation.v1_0.route_manager import RouteManager
from ..protocols.endorse_transaction.v1_0.manager import (
    TransactionManager,
    TransactionManagerError,
)
from ..protocols.endorse_transaction.v1_0.util import (
    get_endorser_connection_id,
    is_author_role,
)
from ..resolver.base import ResolverError
from ..storage.base import BaseStorage
from ..storage.error import StorageError, StorageNotFoundError
from ..storage.record import StorageRecord
from ..storage.type import RECORD_TYPE_ACAPY_UPGRADING
from ..wallet.jwt import jwt_sign, jwt_verify
from ..wallet.sd_jwt import sd_jwt_sign, sd_jwt_verify
from .anoncreds_upgrade import (
    UPGRADING_RECORD_IN_PROGRESS,
    upgrade_wallet_to_anoncreds_if_requested,
)
from .base import BaseWallet
from .did_info import DIDInfo
from .did_method import (
    INDY,
    KEY,
    PEER2,
    PEER4,
    SOV,
    DIDMethod,
    DIDMethods,
    HolderDefinedDid,
)
from .did_posture import DIDPosture
from .error import WalletError, WalletNotFoundError
from .key_type import BLS12381G2, ED25519, P256, KeyTypes
from .singletons import UpgradeInProgressSingleton
from .util import EVENT_LISTENER_PATTERN

LOGGER = logging.getLogger(__name__)

WALLET_TAG_TITLE = "wallet"
UPGRADE_TAG_TITLE = "AnonCreds - Wallet Upgrade"


class WalletModuleResponseSchema(OpenAPISchema):
    """Response schema for Wallet Module."""


class DIDSchema(OpenAPISchema):
    """Result schema for a DID."""

    did = fields.Str(
        required=True,
        validate=GENERIC_DID_VALIDATE,
        metadata={"description": "DID of interest", "example": GENERIC_DID_EXAMPLE},
    )
    verkey = fields.Str(
        required=True,
        validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
        metadata={
            "description": "Public verification key",
            "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
        },
    )
    posture = fields.Str(
        required=True,
        validate=DID_POSTURE_VALIDATE,
        metadata={
            "description": (
                "Whether DID is current public DID, posted to ledger but not current"
                " public DID, or local to the wallet"
            ),
            "example": DID_POSTURE_EXAMPLE,
        },
    )
    method = fields.Str(
        required=True,
        metadata={
            "description": "Did method associated with the DID",
            "example": SOV.method_name,
        },
    )
    key_type = fields.Str(
        required=True,
        validate=validate.OneOf([ED25519.key_type, BLS12381G2.key_type, P256.key_type]),
        metadata={
            "description": "Key type associated with the DID",
            "example": ED25519.key_type,
        },
    )
    metadata = fields.Dict(
        required=False,
        metadata={"description": "Additional metadata associated with the DID"},
    )


class DIDResultSchema(OpenAPISchema):
    """Result schema for a DID."""

    result = fields.Nested(DIDSchema())


class DIDListSchema(OpenAPISchema):
    """Result schema for connection list."""

    results = fields.List(
        fields.Nested(DIDSchema()), metadata={"description": "DID list"}
    )


class DIDEndpointWithTypeSchema(OpenAPISchema):
    """Request schema to set DID endpoint of particular type."""

    did = fields.Str(
        required=True,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "DID of interest", "example": INDY_DID_EXAMPLE},
    )
    endpoint = fields.Str(
        required=False,
        validate=ENDPOINT_VALIDATE,
        metadata={
            "description": "Endpoint to set (omit to delete)",
            "example": ENDPOINT_EXAMPLE,
        },
    )
    endpoint_type = fields.Str(
        required=False,
        validate=ENDPOINT_TYPE_VALIDATE,
        metadata={
            "description": (
                f"Endpoint type to set (default '{EndpointType.ENDPOINT.w3c}'); affects"
                " only public or posted DIDs"
            ),
            "example": ENDPOINT_TYPE_EXAMPLE,
        },
    )
    mediation_id = fields.Str(
        required=False,
        validate=UUID4_VALIDATE,
        metadata={
            "description": "Mediation ID to use for endpoint information.",
            "example": UUID4_EXAMPLE,
        },
    )


class JWSCreateSchema(OpenAPISchema):
    """Request schema to create a jws with a particular DID."""

    headers = fields.Dict()
    payload = fields.Dict(required=True)
    did = fields.Str(
        required=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={"description": "DID of interest", "example": GENERIC_DID_EXAMPLE},
    )
    verification_method = fields.Str(
        data_key="verificationMethod",
        required=False,
        validate=Uri(),
        metadata={
            "description": "Information used for proof verification",
            "example": (
                "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg34"
                "2Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
            ),
        },
    )


class SDJWSCreateSchema(JWSCreateSchema):
    """Request schema to create an sd-jws with a particular DID."""

    non_sd_list = fields.List(
        fields.Str(
            required=False,
            validate=NON_SD_LIST_VALIDATE,
            metadata={"example": NON_SD_LIST_EXAMPLE},
        )
    )


class JWSVerifySchema(OpenAPISchema):
    """Request schema to verify a jws created from a DID."""

    jwt = fields.Str(validate=JWT_VALIDATE, metadata={"example": JWT_EXAMPLE})


class SDJWSVerifySchema(OpenAPISchema):
    """Request schema to verify an sd-jws created from a DID."""

    sd_jwt = fields.Str(validate=SD_JWT_VALIDATE, metadata={"example": SD_JWT_EXAMPLE})


class JWSVerifyResponseSchema(OpenAPISchema):
    """Response schema for JWT verification result."""

    valid = fields.Bool(required=True)
    error = fields.Str(required=False, metadata={"description": "Error text"})
    kid = fields.Str(required=True, metadata={"description": "kid of signer"})
    headers = fields.Dict(
        required=True, metadata={"description": "Headers from verified JWT."}
    )
    payload = fields.Dict(
        required=True, metadata={"description": "Payload from verified JWT"}
    )


class SDJWSVerifyResponseSchema(JWSVerifyResponseSchema):
    """Response schema for SD-JWT verification result."""

    disclosures = fields.List(
        fields.List(StrOrDictField()),
        metadata={
            "description": "Disclosure arrays associated with the SD-JWT",
            "example": [
                ["fx1iT_mETjGiC-JzRARnVg", "name", "Alice"],
                [
                    "n4-t3mlh8jSS6yMIT7QHnA",
                    "street_address",
                    {"_sd": ["kLZrLK7enwfqeOzJ9-Ss88YS3mhjOAEk9lr_ix2Heng"]},
                ],
            ],
        },
    )


class DIDEndpointSchema(OpenAPISchema):
    """Request schema to set DID endpoint; response schema to get DID endpoint."""

    did = fields.Str(
        required=True,
        validate=INDY_DID_VALIDATE,
        metadata={"description": "DID of interest", "example": INDY_DID_EXAMPLE},
    )
    endpoint = fields.Str(
        required=False,
        validate=ENDPOINT_VALIDATE,
        metadata={
            "description": "Endpoint to set (omit to delete)",
            "example": ENDPOINT_EXAMPLE,
        },
    )


class DIDListQueryStringSchema(OpenAPISchema):
    """Parameters and validators for DID list request query string."""

    did = fields.Str(
        required=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={"description": "DID of interest", "example": GENERIC_DID_EXAMPLE},
    )
    verkey = fields.Str(
        required=False,
        validate=RAW_ED25519_2018_PUBLIC_KEY_VALIDATE,
        metadata={
            "description": "Verification key of interest",
            "example": RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE,
        },
    )
    posture = fields.Str(
        required=False,
        validate=DID_POSTURE_VALIDATE,
        metadata={
            "description": (
                "Whether DID is current public DID, posted to ledger but current public"
                " DID, or local to the wallet"
            ),
            "example": DID_POSTURE_EXAMPLE,
        },
    )
    method = fields.Str(
        required=False,
        metadata={
            "example": KEY.method_name,
            "description": (
                "DID method to query for. e.g. sov to only fetch indy/sov DIDs"
            ),
        },
    )
    key_type = fields.Str(
        required=False,
        validate=validate.OneOf([ED25519.key_type, BLS12381G2.key_type, P256.key_type]),
        metadata={"example": ED25519.key_type, "description": "Key type to query for."},
    )


class DIDQueryStringSchema(OpenAPISchema):
    """Parameters and validators for set public DID request query string."""

    did = fields.Str(
        required=True,
        validate=GENERIC_DID_VALIDATE,
        metadata={"description": "DID of interest", "example": GENERIC_DID_EXAMPLE},
    )


class DIDCreateOptionsSchema(OpenAPISchema):
    """Parameters and validators for create DID options."""

    key_type = fields.Str(
        required=True,
        validate=validate.OneOf([ED25519.key_type, BLS12381G2.key_type, P256.key_type]),
        metadata={
            "example": ED25519.key_type,
            "description": (
                "Key type to use for the DID keypair. "
                + "Validated with the chosen DID method's supported key types."
            ),
        },
    )

    did = fields.Str(
        required=False,
        validate=GENERIC_DID_VALIDATE,
        metadata={
            "description": (
                "Specify final value of the did (including did:<method>: prefix)"
                + "if the method supports or requires so."
            ),
            "example": GENERIC_DID_EXAMPLE,
        },
    )


class DIDCreateSchema(OpenAPISchema):
    """Parameters and validators for create DID endpoint."""

    method = fields.Str(
        required=False,
        dump_default=SOV.method_name,
        metadata={
            "example": SOV.method_name,
            "description": (
                "Method for the requested DID."
                + "Supported methods are 'key', 'sov', and any other registered method."
            ),
        },
    )

    options = fields.Nested(
        DIDCreateOptionsSchema,
        required=False,
        metadata={
            "description": (
                "To define a key type and/or a did depending on chosen DID method."
            )
        },
    )

    seed = fields.Str(
        required=False,
        metadata={
            "description": (
                "Optional seed to use for DID, Must be enabled in configuration before"
                " use."
            ),
            "example": "000000000000000000000000Trustee1",
        },
    )


class CreateAttribTxnForEndorserOptionSchema(OpenAPISchema):
    """Class for user to input whether to create a transaction for endorser or not."""

    create_transaction_for_endorser = fields.Boolean(
        required=False,
        metadata={"description": "Create Transaction For Endorser's signature"},
    )


class AttribConnIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking connection id."""

    conn_id = fields.Str(
        required=False, metadata={"description": "Connection identifier"}
    )


class MediationIDSchema(OpenAPISchema):
    """Class for user to optionally input a mediation_id."""

    mediation_id = fields.Str(
        required=False, metadata={"description": "Mediation identifier"}
    )


def format_did_info(info: DIDInfo):
    """Serialize a DIDInfo object."""
    if info:
        return {
            "did": info.did,
            "verkey": info.verkey,
            "posture": DIDPosture.get(info.metadata).moniker,
            "key_type": info.key_type.key_type,
            "method": info.method.method_name,
            "metadata": info.metadata,
        }


@docs(tags=[WALLET_TAG_TITLE], summary="List wallet DIDs")
@querystring_schema(DIDListQueryStringSchema())
@response_schema(DIDListSchema, 200, description="")
@tenant_authentication
async def wallet_did_list(request: web.BaseRequest):
    """Request handler for searching wallet DIDs.

    Args:
        request: aiohttp request object

    Returns:
        The DID list response

    """
    context: AdminRequestContext = request["context"]
    filter_did = request.query.get("did")
    filter_verkey = request.query.get("verkey")
    filter_posture = DIDPosture.get(request.query.get("posture"))
    results = []
    async with context.session() as session:
        did_methods: DIDMethods = session.inject(DIDMethods)
        filter_method: DIDMethod | None = did_methods.from_method(
            request.query.get("method")
        )
        key_types = session.inject(KeyTypes)
        filter_key_type = key_types.from_key_type(request.query.get("key_type", ""))
        wallet: BaseWallet | None = session.inject_or(BaseWallet)
        if not wallet:
            raise web.HTTPForbidden(reason="No wallet available")
        if filter_posture is DIDPosture.PUBLIC:
            public_did_info = await wallet.get_public_did()
            if (
                public_did_info
                and (not filter_verkey or public_did_info.verkey == filter_verkey)
                and (not filter_did or public_did_info.did == filter_did)
                and (not filter_method or public_did_info.method == filter_method)
                and (not filter_key_type or public_did_info.key_type == filter_key_type)
            ):
                results.append(format_did_info(public_did_info))
        elif filter_posture is DIDPosture.POSTED:
            results = []
            posted_did_infos = await wallet.get_posted_dids()
            for info in posted_did_infos:
                if (
                    (not filter_verkey or info.verkey == filter_verkey)
                    and (not filter_did or info.did == filter_did)
                    and (not filter_method or info.method == filter_method)
                    and (not filter_key_type or info.key_type == filter_key_type)
                ):
                    results.append(format_did_info(info))
        elif filter_did:
            try:
                info = await wallet.get_local_did(filter_did)
            except WalletError:
                # badly formatted DID or record not found
                info = None
            if (
                info
                and (not filter_verkey or info.verkey == filter_verkey)
                and (not filter_method or info.method == filter_method)
                and (not filter_key_type or info.key_type == filter_key_type)
                and (
                    filter_posture is None
                    or (
                        filter_posture is DIDPosture.WALLET_ONLY
                        and not info.metadata.get("posted")
                    )
                )
            ):
                results.append(format_did_info(info))
        elif filter_verkey:
            try:
                info = await wallet.get_local_did_for_verkey(filter_verkey)
            except WalletError:
                info = None
            if (
                info
                and (not filter_method or info.method == filter_method)
                and (not filter_key_type or info.key_type == filter_key_type)
                and (
                    filter_posture is None
                    or (
                        filter_posture is DIDPosture.WALLET_ONLY
                        and not info.metadata.get("posted")
                    )
                )
            ):
                results.append(format_did_info(info))
        else:
            dids = await wallet.get_local_dids()
            results = [
                format_did_info(info)
                for info in dids
                if (
                    filter_posture is None
                    or DIDPosture.get(info.metadata) is DIDPosture.WALLET_ONLY
                )
                and (not filter_method or info.method == filter_method)
                and (not filter_key_type or info.key_type == filter_key_type)
            ]

    results.sort(key=lambda info: (DIDPosture.get(info["posture"]).ordinal, info["did"]))

    return web.json_response({"results": results})


@docs(tags=[WALLET_TAG_TITLE], summary="Create a local DID")
@request_schema(DIDCreateSchema())
@response_schema(DIDResultSchema, 200, description="")
@tenant_authentication
async def wallet_create_did(request: web.BaseRequest):
    """Request handler for creating a new local DID in the wallet.

    Args:
        request: aiohttp request object

    Returns:
        The DID info

    """
    context: AdminRequestContext = request["context"]

    try:
        body = await request.json()
    except Exception:
        body = {}

    # set default method and key type for backwards compat

    seed = body.get("seed") or None
    if seed and not context.settings.get("wallet.allow_insecure_seed"):
        raise web.HTTPBadRequest(reason="Seed support is not enabled")
    info = None
    async with context.session() as session:
        did_methods = session.inject(DIDMethods)

        method = did_methods.from_method(body.get("method", "sov"))
        if not method:
            raise web.HTTPForbidden(
                reason=f"method {body.get('method')} is not supported by the agent."
            )

        # Don't support Indy DID method from this endpoint
        if method.method_name == INDY.method_name:
            raise web.HTTPBadRequest(
                reason="Indy did method is supported from /did/indy/create endpoint."
            )

        key_types = session.inject(KeyTypes)
        # set default method and key type for backwards compat
        key_type = (
            key_types.from_key_type(body.get("options", {}).get("key_type", ""))
            or ED25519
        )
        if not method.supports_key_type(key_type):
            raise web.HTTPForbidden(
                reason=(
                    f"method {method.method_name} does not"
                    f" support key type {key_type.key_type}"
                )
            )

        did = body.get("options", {}).get("did")
        if method.holder_defined_did() == HolderDefinedDid.NO and did:
            raise web.HTTPForbidden(
                reason=f"method {method.method_name} does not support user-defined DIDs"
            )
        elif method.holder_defined_did() == HolderDefinedDid.REQUIRED and not did:
            raise web.HTTPBadRequest(
                reason=f"method {method.method_name} requires a user-defined DIDs"
            )

        wallet = session.inject_or(BaseWallet)
        if not wallet:
            raise web.HTTPForbidden(reason="No wallet available")
        try:
            is_did_peer_2 = method.method_name == PEER2.method_name
            is_did_peer_4 = method.method_name == PEER4.method_name
            if is_did_peer_2 or is_did_peer_4:
                base_conn_mgr = BaseConnectionManager(context.profile)

                options = body.get("options", {})

                connection_id = options.get("conn_id")
                my_endpoint = options.get("endpoint")
                mediation_id = options.get("mediation_id")

                # FIXME:
                # This logic is duplicated in BaseConnectionManager
                # It should be refactored into one method.
                ###################################################

                my_endpoints = []
                if my_endpoint:
                    my_endpoints = [my_endpoint]
                else:
                    default_endpoint = context.profile.settings.get("default_endpoint")
                    if default_endpoint:
                        my_endpoints.append(default_endpoint)
                    my_endpoints.extend(
                        context.profile.settings.get("additional_endpoints", [])
                    )

                mediation_records = []
                if connection_id:
                    conn_rec = await ConnRecord.retrieve_by_id(session, connection_id)
                    mediation_records = await base_conn_mgr._route_manager.mediation_records_for_connection(  # noqa: E501
                        context.profile,
                        conn_rec,
                        mediation_id,
                        or_default=True,
                    )

                ###################################################

                info = (
                    await base_conn_mgr.create_did_peer_2(my_endpoints, mediation_records)
                    if is_did_peer_2
                    else await base_conn_mgr.create_did_peer_4(
                        my_endpoints, mediation_records
                    )
                )
            else:
                info = await wallet.create_local_did(
                    method=method, key_type=key_type, seed=seed, did=did
                )

        except WalletError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"result": format_did_info(info)})


@docs(tags=[WALLET_TAG_TITLE], summary="Fetch the current public DID")
@response_schema(DIDResultSchema, 200, description="")
@tenant_authentication
async def wallet_get_public_did(request: web.BaseRequest):
    """Request handler for fetching the current public DID.

    Args:
        request: aiohttp request object

    Returns:
        The DID info

    """
    context: AdminRequestContext = request["context"]
    info = None
    async with context.session() as session:
        wallet = session.inject_or(BaseWallet)
        if not wallet:
            raise web.HTTPForbidden(reason="No wallet available")
        try:
            info = await wallet.get_public_did()
        except WalletError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"result": format_did_info(info)})


@docs(tags=[WALLET_TAG_TITLE], summary="Assign the current public DID")
@querystring_schema(DIDQueryStringSchema())
@querystring_schema(CreateAttribTxnForEndorserOptionSchema())
@querystring_schema(AttribConnIdMatchInfoSchema())
@querystring_schema(MediationIDSchema())
@response_schema(DIDResultSchema, 200, description="")
@tenant_authentication
async def wallet_set_public_did(request: web.BaseRequest):
    """Request handler for setting the current public DID.

    Args:
        request: aiohttp request object

    Returns:
        The updated DID info

    """
    context: AdminRequestContext = request["context"]

    outbound_handler = request["outbound_message_router"]

    create_transaction_for_endorser = json.loads(
        request.query.get("create_transaction_for_endorser", "false")
    )
    write_ledger = not create_transaction_for_endorser
    connection_id = request.query.get("conn_id")
    attrib_def = None

    # check if we need to endorse
    if is_author_role(context.profile):
        # authors cannot write to the ledger
        write_ledger = False
        create_transaction_for_endorser = True
        if not connection_id:
            # author has not provided a connection id, so determine which to use
            connection_id = await get_endorser_connection_id(context.profile)
            if not connection_id:
                raise web.HTTPBadRequest(reason="No endorser connection found")

    async with context.session() as session:
        wallet = session.inject_or(BaseWallet)
        if not wallet:
            raise web.HTTPForbidden(reason="No wallet available")
    did = request.query.get("did")
    if not did:
        raise web.HTTPBadRequest(reason="Request query must include DID")

    info: Optional[DIDInfo] = None

    mediation_id = request.query.get("mediation_id")
    profile = context.profile
    route_manager = profile.inject(RouteManager)
    mediation_record = await route_manager.mediation_record_if_id(
        profile=profile, mediation_id=mediation_id, or_default=True
    )

    routing_keys, mediator_endpoint = await route_manager.routing_info(
        profile,
        mediation_record,
    )

    try:
        info, attrib_def = await promote_wallet_public_did(
            context,
            did,
            write_ledger=write_ledger,
            connection_id=connection_id,
            routing_keys=routing_keys,
            mediator_endpoint=mediator_endpoint,
        )
    except LookupError as err:
        raise web.HTTPNotFound(reason=str(err)) from err
    except PermissionError as err:
        raise web.HTTPForbidden(reason=str(err)) from err
    except WalletNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (LedgerError, WalletError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    if not create_transaction_for_endorser:
        return web.json_response({"result": format_did_info(info)})
    else:
        # DID is already posted to ledger
        if not attrib_def:
            return web.json_response({"result": format_did_info(info)})

        transaction_mgr = TransactionManager(context.profile)
        try:
            transaction = await transaction_mgr.create_record(
                messages_attach=attrib_def["signed_txn"], connection_id=connection_id
            )
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        # if auto-request, send the request to the endorser
        if context.settings.get_value("endorser.auto_request"):
            try:
                transaction, transaction_request = await transaction_mgr.create_request(
                    transaction=transaction,
                    # TODO see if we need to parametrize these params
                    # expires_time=expires_time,
                )
            except (StorageError, TransactionManagerError) as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

            await outbound_handler(transaction_request, connection_id=connection_id)

        return web.json_response({"txn": transaction.serialize()})


async def promote_wallet_public_did(
    context: Union[AdminRequestContext, InjectionContext],
    did: str,
    write_ledger: bool = False,
    profile: Optional[Profile] = None,
    connection_id: Optional[str] = None,
    routing_keys: Optional[List[str]] = None,
    mediator_endpoint: Optional[str] = None,
) -> Tuple[DIDInfo, Optional[dict]]:
    """Promote supplied DID to the wallet public DID."""
    LOGGER.debug("Starting promotion of DID %s to wallet public DID", did)
    info: Optional[DIDInfo] = None
    endorser_did = None

    is_indy_did = bool(IndyDID.PATTERN.match(did))
    # write only Indy DID
    write_ledger = is_indy_did and write_ledger
    is_ctx_admin_request = True
    if isinstance(context, InjectionContext):
        is_ctx_admin_request = False
        if not profile:
            LOGGER.error("InjectionContext provided without profile")
            raise web.HTTPForbidden(
                reason=(
                    "InjectionContext is provided but no profile is provided. "
                    "InjectionContext does not have profile attribute but "
                    "AdminRequestContext does."
                )
            )
    ledger = (
        context.profile.inject_or(BaseLedger)
        if is_ctx_admin_request
        else profile.inject_or(BaseLedger)
    )

    if is_indy_did:
        if not ledger:
            reason = "No ledger available"
            if not context.settings.get_value("wallet.type"):
                reason += ": missing wallet-type?"
            LOGGER.info("Cannot promote DID %s to public DID: %s", did, reason)
            raise PermissionError(reason)

        async with ledger:
            if not await ledger.get_key_for_did(did):
                LOGGER.info("Cannot promote DID %s; it is not posted to the ledger", did)
                raise LookupError(f"DID {did} is not posted to the ledger")

        is_author_profile = (
            is_author_role(context.profile)
            if is_ctx_admin_request
            else is_author_role(profile)
        )

        # check if we need to endorse
        if is_author_profile:
            # authors cannot write to the ledger
            write_ledger = False

            LOGGER.debug("No connection id provided; determining which to use")
            if not connection_id:
                connection_id = (
                    await get_endorser_connection_id(context.profile)
                    if is_ctx_admin_request
                    else await get_endorser_connection_id(profile)
                )
            if not connection_id:
                LOGGER.info("Cannot promote DID %s; no endorser connection found", did)
                raise web.HTTPBadRequest(reason="No endorser connection found")
        if not write_ledger:
            async with (
                context.session() if is_ctx_admin_request else profile.session()
            ) as session:
                try:
                    connection_record = await ConnRecord.retrieve_by_id(
                        session, connection_id
                    )
                except StorageNotFoundError as err:
                    LOGGER.info("Connection record not found: %s", err.roll_up)
                    raise web.HTTPNotFound(reason=err.roll_up) from err
                except BaseModelError as err:
                    LOGGER.error("Base model error: %s", err.roll_up)
                    raise web.HTTPBadRequest(reason=err.roll_up) from err
                endorser_info = await connection_record.metadata_get(
                    session, "endorser_info"
                )

            if not endorser_info:
                LOGGER.info(
                    "Cannot promote %s; endorser info not set up in connection metadata",
                    did,
                )
                raise web.HTTPForbidden(
                    reason=(
                        "Endorser Info is not set up in "
                        "connection metadata for this connection record"
                    )
                )
            if "endorser_did" not in endorser_info.keys():
                LOGGER.info(
                    'Cannot promote DID %s; "endorser_did" not set in "endorser_info"',
                    did,
                )
                raise web.HTTPForbidden(
                    reason=(
                        ' "endorser_did" is not set in "endorser_info"'
                        " in connection metadata for this connection record"
                    )
                )
            endorser_did = endorser_info["endorser_did"]
            LOGGER.debug("Endorser DID %s found in connection metadata", endorser_did)

    did_info: Optional[DIDInfo] = None
    attrib_def = None
    async with (
        context.session() if is_ctx_admin_request else profile.session()
    ) as session:
        wallet = session.inject(BaseWallet)
        did_info = await wallet.get_local_did(did)
        info = await wallet.set_public_did(did_info)
        LOGGER.info("DID %s set as public DID", info.did)

        if info:
            # Publish endpoint if necessary
            endpoint = did_info.metadata.get("endpoint")

            if is_indy_did and not endpoint:
                endpoint = mediator_endpoint or context.settings.get("default_endpoint")
                LOGGER.debug("Setting endpoint for DID %s to %s", info.did, endpoint)
                attrib_def = await wallet.set_did_endpoint(
                    info.did,
                    endpoint,
                    ledger,
                    write_ledger=write_ledger,
                    endorser_did=endorser_did,
                    routing_keys=routing_keys,
                )
                LOGGER.debug("Endpoint set for DID %s: %s", info.did, endpoint)

    if info:
        LOGGER.debug("Routing public DID %s", info.did)
        if is_ctx_admin_request:
            profile = context.profile
        route_manager = profile.inject(RouteManager)
        await route_manager.route_verkey(profile, info.verkey)
        LOGGER.info(
            "Routing set up for public DID %s with verkey %s", info.did, info.verkey
        )

    LOGGER.debug("Completed promotion of DID %s", did)
    return info, attrib_def


@docs(
    tags=[WALLET_TAG_TITLE],
    summary="Update endpoint in wallet and on ledger if posted to it",
)
@request_schema(DIDEndpointWithTypeSchema)
@querystring_schema(CreateAttribTxnForEndorserOptionSchema())
@querystring_schema(AttribConnIdMatchInfoSchema())
@response_schema(WalletModuleResponseSchema(), description="")
@tenant_authentication
async def wallet_set_did_endpoint(request: web.BaseRequest):
    """Request handler for setting an endpoint for a DID.

    Args:
        request: aiohttp request object
    """
    context: AdminRequestContext = request["context"]

    outbound_handler = request["outbound_message_router"]

    body = await request.json()
    did = body["did"]
    endpoint = body.get("endpoint")
    endpoint_type = EndpointType.get(body.get("endpoint_type", EndpointType.ENDPOINT.w3c))
    mediation_id = body.get("mediation_id")

    create_transaction_for_endorser = json.loads(
        request.query.get("create_transaction_for_endorser", "false")
    )
    write_ledger = not create_transaction_for_endorser
    endorser_did = None
    connection_id = request.query.get("conn_id")
    attrib_def = None

    profile = context.profile
    route_manager = profile.inject(RouteManager)
    mediation_record = await route_manager.mediation_record_if_id(
        profile=profile, mediation_id=mediation_id, or_default=True
    )
    routing_keys, mediator_endpoint = await route_manager.routing_info(
        profile,
        mediation_record,
    )
    LOGGER.debug("Mediation info: %s, %s", routing_keys, mediator_endpoint)

    # check if we need to endorse
    if is_author_role(context.profile):
        # authors cannot write to the ledger
        write_ledger = False
        create_transaction_for_endorser = True
        if not connection_id:
            # author has not provided a connection id, so determine which to use
            connection_id = await get_endorser_connection_id(context.profile)
            if not connection_id:
                raise web.HTTPBadRequest(reason="No endorser connection found")

    if not write_ledger:
        try:
            async with context.session() as session:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, connection_id
                )
        except StorageNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except BaseModelError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        async with context.session() as session:
            endorser_info = await connection_record.metadata_get(session, "endorser_info")
        if not endorser_info:
            raise web.HTTPForbidden(
                reason=(
                    "Endorser Info is not set up in "
                    "connection metadata for this connection record"
                )
            )
        if "endorser_did" not in endorser_info.keys():
            raise web.HTTPForbidden(
                reason=(
                    ' "endorser_did" is not set in "endorser_info"'
                    " in connection metadata for this connection record"
                )
            )
        endorser_did = endorser_info["endorser_did"]

    async with context.session() as session:
        wallet = session.inject_or(BaseWallet)
        if not wallet:
            raise web.HTTPForbidden(reason="No wallet available")
        try:
            endpoint = mediator_endpoint or endpoint
            ledger = context.profile.inject_or(BaseLedger)
            attrib_def = await wallet.set_did_endpoint(
                did,
                endpoint,
                ledger,
                endpoint_type,
                write_ledger=write_ledger,
                endorser_did=endorser_did,
                routing_keys=routing_keys,
            )
        except WalletNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except LedgerConfigError as err:
            raise web.HTTPForbidden(reason=err.roll_up) from err
        except (LedgerError, WalletError) as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    if not create_transaction_for_endorser:
        return web.json_response({})
    else:
        transaction_mgr = TransactionManager(context.profile)
        try:
            transaction = await transaction_mgr.create_record(
                messages_attach=attrib_def["signed_txn"], connection_id=connection_id
            )
        except StorageError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

        # if auto-request, send the request to the endorser
        if context.settings.get_value("endorser.auto_request"):
            try:
                transaction, transaction_request = await transaction_mgr.create_request(
                    transaction=transaction,
                    # TODO see if we need to parametrize these params
                    # expires_time=expires_time,
                )
            except (StorageError, TransactionManagerError) as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err

            await outbound_handler(transaction_request, connection_id=connection_id)

        return web.json_response({"txn": transaction.serialize()})


@docs(tags=[WALLET_TAG_TITLE], summary="Create a jws using did keys with a given payload")
@request_schema(JWSCreateSchema)
@response_schema(WalletModuleResponseSchema(), description="")
@tenant_authentication
async def wallet_jwt_sign(request: web.BaseRequest):
    """Request handler for jws creation using did.

    Args:
        request (web.BaseRequest): The HTTP request object.

    Returns:
        web.Response: The HTTP response containing the signed JWS.

    Raises:
        web.HTTPBadRequest: If the provided DID or verification method is invalid.
        web.HTTPNotFound: If the wallet is not found.
        web.HTTPBadRequest: If there is an error with the wallet.

    """
    context: AdminRequestContext = request["context"]
    body = await request.json()
    did = body.get("did")
    verification_method = body.get("verificationMethod")
    headers = body.get("headers", {})
    payload = body.get("payload", {})

    try:
        jws = await jwt_sign(context.profile, headers, payload, did, verification_method)
    except ValueError as err:
        raise web.HTTPBadRequest(reason="Bad did or verification method") from err
    except WalletNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except WalletError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(jws)


@docs(
    tags=[WALLET_TAG_TITLE],
    summary="Create an sd-jws using did keys with a given payload",
)
@request_schema(SDJWSCreateSchema)
@response_schema(WalletModuleResponseSchema(), description="")
@tenant_authentication
async def wallet_sd_jwt_sign(request: web.BaseRequest):
    """Request handler for sd-jws creation using did.

    Args:
        request (web.BaseRequest): The HTTP request object.

    Returns:
        web.Response: The HTTP response object.
            Contains the signed sd-jws.

    Raises:
        web.HTTPBadRequest: If the provided did or verification method is invalid.
        web.HTTPNotFound: If the wallet is not found.
        web.HTTPBadRequest: If there is an error with the wallet.

    """
    context: AdminRequestContext = request["context"]
    body = await request.json()
    did = body.get("did")
    verification_method = body.get("verificationMethod")
    headers = body.get("headers", {})
    payload = body.get("payload", {})
    non_sd_list = body.get("non_sd_list", [])

    try:
        sd_jws = await sd_jwt_sign(
            context.profile, headers, payload, non_sd_list, did, verification_method
        )
    except ValueError as err:
        raise web.HTTPBadRequest(reason="Bad did or verification method") from err
    except WalletNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except WalletError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(sd_jws)


@docs(tags=[WALLET_TAG_TITLE], summary="Verify a jws using did keys with a given JWS")
@request_schema(JWSVerifySchema())
@response_schema(JWSVerifyResponseSchema(), 200, description="")
@tenant_authentication
async def wallet_jwt_verify(request: web.BaseRequest):
    """Request handler for jws validation using did.

    Args:
        request (web.BaseRequest): The HTTP request object.
            "jwt": { ... }

    Returns:
        web.Response: The HTTP response containing the validation result.

    Raises:
        web.HTTPBadRequest: If there is an error with the JWS header or verification
            method.
        web.HTTPNotFound: If there is an error resolving the JWS.

    """
    context: AdminRequestContext = request["context"]
    body = await request.json()
    jwt = body["jwt"]
    try:
        result = await jwt_verify(context.profile, jwt)
    except (BadJWSHeaderError, InvalidVerificationMethod) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    except ResolverError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    return web.json_response(
        {
            "valid": result.valid,
            "headers": result.headers,
            "payload": result.payload,
            "kid": result.kid,
        }
    )


@docs(
    tags=[WALLET_TAG_TITLE],
    summary="Verify an sd-jws using did keys with a given SD-JWS with "
    "optional key binding",
)
@request_schema(SDJWSVerifySchema())
@response_schema(SDJWSVerifyResponseSchema(), 200, description="")
@tenant_authentication
async def wallet_sd_jwt_verify(request: web.BaseRequest):
    """Request handler for sd-jws validation using did.

    Args:
        request: The web request object.
            "sd-jwt": { ... }

    Returns:
        A JSON response containing the result of the sd-jwt verification.

    Raises:
        web.HTTPBadRequest: If there is an error with the JWS header or verification
            method.
        web.HTTPNotFound: If there is an error resolving the verification method.
    """
    context: AdminRequestContext = request["context"]
    body = await request.json()
    sd_jwt = body["sd_jwt"]
    try:
        result = await sd_jwt_verify(context.profile, sd_jwt)
    except (BadJWSHeaderError, InvalidVerificationMethod) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err
    except ResolverError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    return web.json_response(result.serialize())


@docs(tags=[WALLET_TAG_TITLE], summary="Query DID endpoint in wallet")
@querystring_schema(DIDQueryStringSchema())
@response_schema(DIDEndpointSchema, 200, description="")
@tenant_authentication
async def wallet_get_did_endpoint(request: web.BaseRequest):
    """Request handler for getting the current DID endpoint from the wallet.

    Args:
        request: aiohttp request object

    Returns:
        The updated DID info

    """
    context: AdminRequestContext = request["context"]
    async with context.session() as session:
        wallet = session.inject_or(BaseWallet)
        if not wallet:
            raise web.HTTPForbidden(reason="No wallet available")
        did = request.query.get("did")
        if not did:
            raise web.HTTPBadRequest(reason="Request query must include DID")

        try:
            did_info = await wallet.get_local_did(did)
            endpoint = did_info.metadata.get("endpoint")
        except WalletNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except WalletError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"did": did, "endpoint": endpoint})


@docs(
    tags=[WALLET_TAG_TITLE], summary="Rotate keypair for a DID not posted to the ledger"
)
@querystring_schema(DIDQueryStringSchema())
@response_schema(WalletModuleResponseSchema(), description="")
@tenant_authentication
async def wallet_rotate_did_keypair(request: web.BaseRequest):
    """Request handler for rotating local DID keypair.

    Args:
        request: aiohttp request object

    Returns:
        An empty JSON response

    """
    context: AdminRequestContext = request["context"]
    did = request.query.get("did")
    if not did:
        raise web.HTTPBadRequest(reason="Request query must include DID")

    async with context.session() as session:
        wallet = session.inject_or(BaseWallet)
        if not wallet:
            raise web.HTTPForbidden(reason="No wallet available")
        try:
            did_info: Optional[DIDInfo] = None
            did_info = await wallet.get_local_did(did)
            if did_info.metadata.get("posted", False):
                # call from ledger API instead to propagate through ledger NYM transaction
                raise web.HTTPBadRequest(reason=f"DID {did} is posted to the ledger")
            await wallet.rotate_did_keypair_start(did)  # do not take seed over the wire
            await wallet.rotate_did_keypair_apply(did)
        except WalletNotFoundError as err:
            raise web.HTTPNotFound(reason=err.roll_up) from err
        except WalletError as err:
            raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({})


class UpgradeVerificationSchema(OpenAPISchema):
    """Parameters and validators for triggering an upgrade to anoncreds."""

    wallet_name = fields.Str(
        required=True,
        metadata={
            "description": "Name of wallet to upgrade to anoncreds",
            "example": "base-wallet",
        },
    )


class UpgradeResultSchema(OpenAPISchema):
    """Result schema for upgrade."""


@docs(
    tags=[UPGRADE_TAG_TITLE],
    summary="Upgrade the wallet from askar to askar-anoncreds. Be very careful with this!"
    " You cannot go back! See migration guide for more information.",
)
@querystring_schema(UpgradeVerificationSchema())
@response_schema(UpgradeResultSchema(), description="")
@tenant_authentication
async def upgrade_anoncreds(request: web.BaseRequest):
    """Request handler for triggering an upgrade to anoncreds.

    Args:
        request: aiohttp request object

    Returns:
        An empty JSON response

    """
    context: AdminRequestContext = request["context"]
    profile = context.profile

    if profile.settings.get("wallet.name") != request.query.get("wallet_name"):
        raise web.HTTPBadRequest(
            reason="Wallet name parameter does not match the agent which triggered the upgrade"  # noqa: E501
        )

    if profile.settings.get("wallet.type") == "askar-anoncreds":
        raise web.HTTPBadRequest(reason="Wallet type is already anoncreds")

    async with profile.session() as session:
        storage = session.inject(BaseStorage)
        upgrading_record = StorageRecord(
            RECORD_TYPE_ACAPY_UPGRADING,
            UPGRADING_RECORD_IN_PROGRESS,
        )
        await storage.add_record(upgrading_record)
        is_subwallet = context.metadata and "wallet_id" in context.metadata
        asyncio.create_task(
            upgrade_wallet_to_anoncreds_if_requested(profile, is_subwallet)
        )
        UpgradeInProgressSingleton().set_wallet(profile.name)

    return web.json_response(
        {
            "success": True,
            "message": f"Upgrade to anoncreds has been triggered for wallet {profile.name}",  # noqa: E501
        }
    )


def register_events(event_bus: EventBus):
    """Subscribe to any events we need to support."""
    event_bus.subscribe(EVENT_LISTENER_PATTERN, on_register_nym_event)


async def on_register_nym_event(profile: Profile, event: Event):
    """Handle any events we need to support."""

    # after the nym record is written, promote to wallet public DID
    if is_author_role(profile) and profile.context.settings.get_value(
        "endorser.auto_promote_author_did"
    ):
        did = event.payload["did"]
        connection_id = event.payload.get("connection_id")
        try:
            _info, attrib_def = await promote_wallet_public_did(
                context=profile.context,
                did=did,
                connection_id=connection_id,
                profile=profile,
            )
        except Exception as err:
            # log the error, but continue
            LOGGER.exception(
                "Error promoting to public DID: %s",
                err,
            )
            return

        transaction_mgr = TransactionManager(profile)
        try:
            transaction = await transaction_mgr.create_record(
                messages_attach=attrib_def["signed_txn"], connection_id=connection_id
            )
        except StorageError as err:
            # log the error, but continue
            LOGGER.exception(
                "Error accepting endorser invitation/configuring endorser connection: %s",
                err,
            )
            return

        # if auto-request, send the request to the endorser
        if profile.settings.get_value("endorser.auto_request"):
            try:
                transaction, transaction_request = await transaction_mgr.create_request(
                    transaction=transaction,
                    # TODO see if we need to parametrize these params
                    # expires_time=expires_time,
                )
            except (StorageError, TransactionManagerError) as err:
                # log the error, but continue
                LOGGER.exception(
                    "Error creating endorser transaction request: %s",
                    err,
                )

            # TODO not sure how to get outbound_handler in an event ...
            # await outbound_handler(transaction_request, connection_id=connection_id)
            responder = profile.inject_or(BaseResponder)
            if responder:
                await responder.send(
                    transaction_request,
                    connection_id=connection_id,
                )
            else:
                LOGGER.warning(
                    "Configuration has no BaseResponder: cannot update "
                    "ATTRIB record on DID: %s",
                    did,
                )


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get("/wallet/did", wallet_did_list, allow_head=False),
            web.post("/wallet/did/create", wallet_create_did),
            web.get("/wallet/did/public", wallet_get_public_did, allow_head=False),
            web.post("/wallet/did/public", wallet_set_public_did),
            web.post("/wallet/set-did-endpoint", wallet_set_did_endpoint),
            web.post("/wallet/jwt/sign", wallet_jwt_sign),
            web.post("/wallet/jwt/verify", wallet_jwt_verify),
            web.post("/wallet/sd-jwt/sign", wallet_sd_jwt_sign),
            web.post("/wallet/sd-jwt/verify", wallet_sd_jwt_verify),
            web.get(
                "/wallet/get-did-endpoint", wallet_get_did_endpoint, allow_head=False
            ),
            web.patch("/wallet/did/local/rotate-keypair", wallet_rotate_did_keypair),
            web.post("/anoncreds/wallet/upgrade", upgrade_anoncreds),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": WALLET_TAG_TITLE,
            "description": "DID and tag policy management",
            "externalDocs": {
                "description": "Design",
                "url": (
                    "https://github.com/hyperledger/indy-sdk/tree/"
                    "master/docs/design/003-wallet-storage"
                ),
            },
        }
    )
    app._state["swagger_dict"]["tags"].append(
        {
            "name": UPGRADE_TAG_TITLE,
            "description": "AnonCreds wallet upgrade",
            "externalDocs": {
                "description": "Specification",
                "url": "https://hyperledger.github.io/anoncreds-spec",
            },
        }
    )



"""Cred request artifacts to attach to RFC 453 messages."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .......messaging.models.base import BaseModel, BaseModelSchema


class VCDIIndyCredential(BaseModel):
    """VCDI Indy credential."""

    class Meta:
        """VCDI Indy credential metadata."""

        schema_class = "VCDIIndyCredentialSchema"

    def __init__(
        self,
        credential: Optional[dict] = None,
        **kwargs,
    ):
        """Initialize vcdi cred abstract object.

        Args:
            credential: credential object
            kwargs: additional keyword arguments
        """
        super().__init__(**kwargs)
        self.credential = credential


class VCDIIndyCredentialSchema(BaseModelSchema):
    """VCDI Indy credential schema."""

    class Meta:
        """VCDI Indy credential schemametadata."""

        model_class = VCDIIndyCredential
        unknown = EXCLUDE

    credential = fields.Dict(
        fields.Str(), required=True, metadata={"description": "", "example": ""}
    )



from unittest import IsolatedAsyncioTestCase, TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import TRANSACTION_ACKNOWLEDGEMENT
from ..transaction_acknowledgement import TransactionAcknowledgement


class TestConfig:
    test_thread_id = "3fa85f64-5717-4562-b3fc-2c963f66afa6"


class TestTransactionAcknowledgement(TestCase, TestConfig):
    def setUp(self):
        self.transaction_acknowledgement = TransactionAcknowledgement(
            thread_id=self.test_thread_id
        )

    def test_init(self):
        """Test initialization."""
        assert self.transaction_acknowledgement.thread_id == self.test_thread_id

    def test_type(self):
        """Test type."""
        assert self.transaction_acknowledgement._type == DIDCommPrefix.qualify_current(
            TRANSACTION_ACKNOWLEDGEMENT
        )

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "transaction_acknowledgement.TransactionAcknowledgementSchema.load"
    )
    def test_deserialize(self, mock_transaction_acknowledgement_schema_load):
        """
        Test deserialization.
        """
        obj = self.transaction_acknowledgement

        transaction_acknowledgement = TransactionAcknowledgement.deserialize(obj)
        mock_transaction_acknowledgement_schema_load.assert_called_once_with(obj)

        assert (
            transaction_acknowledgement
            is mock_transaction_acknowledgement_schema_load.return_value
        )

    @mock.patch(
        "acapy_agent.protocols.endorse_transaction.v1_0.messages."
        "transaction_acknowledgement.TransactionAcknowledgementSchema.dump"
    )
    def test_serialize(self, mock_transaction_acknowledgement_schema_dump):
        """
        Test serialization.
        """
        transaction_acknowledgement_dict = self.transaction_acknowledgement.serialize()
        mock_transaction_acknowledgement_schema_dump.assert_called_once_with(
            self.transaction_acknowledgement
        )

        assert (
            transaction_acknowledgement_dict
            is mock_transaction_acknowledgement_schema_dump.return_value
        )


class TestTransactionAcknowledgementSchema(IsolatedAsyncioTestCase, TestConfig):
    """Test transaction acknowledgement schema."""

    async def test_make_model(self):
        transaction_acknowledgement = TransactionAcknowledgement(
            thread_id=self.test_thread_id
        )
        data = transaction_acknowledgement.serialize()
        model_instance = TransactionAcknowledgement.deserialize(data)
        assert type(model_instance) is type(transaction_acknowledgement)



"""Represents a transaction acknowledgement message."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.valid import UUID4_EXAMPLE
from ....notification.v1_0.messages.ack import V10Ack, V10AckSchema
from ..message_types import PROTOCOL_PACKAGE, TRANSACTION_ACKNOWLEDGEMENT

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers"
    ".transaction_acknowledgement_handler.TransactionAcknowledgementHandler"
)


class TransactionAcknowledgement(V10Ack):
    """Class representing a transaction acknowledgement message."""

    class Meta:
        """Metadata for a transaction acknowledgement message."""

        handler_class = HANDLER_CLASS
        message_type = TRANSACTION_ACKNOWLEDGEMENT
        schema_class = "TransactionAcknowledgementSchema"

    def __init__(
        self,
        *,
        thread_id: Optional[str] = None,
        ledger_response: Optional[dict] = None,
        **kwargs,
    ):
        """Initialize a transaction acknowledgement object.

        Args:
            thread_id: Thread id of transaction record
            ledger_response: Response from the ledger
            kwargs: Additional keyword arguments for the message
        """
        super().__init__(**kwargs)
        self.thread_id = thread_id
        self.ledger_response = ledger_response


class TransactionAcknowledgementSchema(V10AckSchema):
    """Transaction Acknowledgement schema class."""

    class Meta:
        """Transaction Acknowledgement metadata."""

        model_class = TransactionAcknowledgement
        unknown = EXCLUDE

    thread_id = fields.Str(required=True, metadata={"example": UUID4_EXAMPLE})
    ledger_response = fields.Dict(required=False)



DID_ED25519_z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th = {
    "@context": ["https://www.w3.org/ns/did/v1"],
    "id": "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th",
    "verificationMethod": [
        {
            "id": "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th#z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th",
            "type": "Ed25519VerificationKey2018",
            "controller": "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th",
            "publicKeyBase58": "8HH5gYEeNc3z7PYXmd54d4x6qAfCNrqQqEB3nS7Zfu7K",
        }
    ],
    "authentication": [
        "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th#z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th"
    ],
    "assertionMethod": [
        "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th#z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th"
    ],
    "capabilityDelegation": [
        "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th#z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th"
    ],
    "capabilityInvocation": [
        "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th#z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th"
    ],
    "keyAgreement": [
        {
            "id": "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th#z6LShpNhGwSupbB7zjuivH156vhLJBDDzmQtA4BY9S94pe1K",
            "type": "X25519KeyAgreementKey2019",
            "controller": "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th",
            "publicKeyBase58": "79CXkde3j8TNuMXxPdV7nLUrT2g7JAEjH5TreyVY7GEZ",
        }
    ],
}

DID_X25519_z6LShLeXRTzevtwcfehaGEzCMyL3bNsAeKCwcqwJxyCo63yE = {
    "@context": "https://www.w3.org/ns/did/v1",
    "id": "did:key:z6LShLeXRTzevtwcfehaGEzCMyL3bNsAeKCwcqwJxyCo63yE",
    "verificationMethod": [
        {
            "id": "did:key:z6LShLeXRTzevtwcfehaGEzCMyL3bNsAeKCwcqwJxyCo63yE#z6LShLeXRTzevtwcfehaGEzCMyL3bNsAeKCwcqwJxyCo63yE",
            "type": "X25519KeyAgreementKey2019",
            "controller": "did:key:z6LShLeXRTzevtwcfehaGEzCMyL3bNsAeKCwcqwJxyCo63yE",
            "publicKeyBase58": "6fUMuABnqSDsaGKojbUF3P7ZkEL3wi2njsDdUWZGNgCU",
        }
    ],
    "authentication": [],
    "assertionMethod": [],
    "capabilityDelegation": [],
    "capabilityInvocation": [],
    "keyAgreement": [
        "did:key:z6LShLeXRTzevtwcfehaGEzCMyL3bNsAeKCwcqwJxyCo63yE#z6LShLeXRTzevtwcfehaGEzCMyL3bNsAeKCwcqwJxyCo63yE"
    ],
}

DID_P256_zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169 = {
    "@context": ["https://www.w3.org/ns/did/v1", "https://w3id.org/security/multikey/v1"],
    "id": "did:key:zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169",
    "verificationMethod": [
        {
            "id": "did:key:zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169#zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169",
            "type": "Multikey",
            "controller": "did:key:zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169",
            "publicKeyMultibase": "zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169",
        }
    ],
    "authentication": [
        "did:key:zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169#zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169"
    ],
    "assertionMethod": [
        "did:key:zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169#zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169"
    ],
    "capabilityDelegation": [
        "did:key:zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169#zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169"
    ],
    "capabilityInvocation": [
        "did:key:zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169#zDnaerDaTF5BXEavCrfRZEk316dpbLsfPDZ3WJ5hRTPFU2169"
    ],
    "keyAgreement": [],
}

DID_BLS12381G2_zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT = {
    "@context": ["https://www.w3.org/ns/did/v1"],
    "id": "did:key:zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT",
    "verificationMethod": [
        {
            "id": "did:key:zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT#zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT",
            "type": "Bls12381G2Key2020",
            "controller": "did:key:zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT",
            "publicKeyBase58": "mxE4sHTpbPcmxNviRVR9r7D2taXcNyVJmf9TBUFS1gRt3j3Ej9Seo59GQeCzYwbQgDrfWCwEJvmBwjLvheAky5N2NqFVzk4kuq3S8g4Fmekai4P622vHqWjFrsioYYDqhf9",
        }
    ],
    "authentication": [
        "did:key:zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT#zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT"
    ],
    "assertionMethod": [
        "did:key:zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT#zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT"
    ],
    "capabilityDelegation": [
        "did:key:zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT#zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT"
    ],
    "capabilityInvocation": [
        "did:key:zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT#zUC71nmwvy83x1UzNKbZbS7N9QZx8rqpQx3Ee3jGfKiEkZngTKzsRoqobX6wZdZF5F93pSGYYco3gpK9tc53ruWUo2tkBB9bxPCFBUjq2th8FbtT4xih6y6Q1K9EL4Th86NiCGT"
    ],
    "keyAgreement": [],
}

DID_BLS12381G1_z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA = {
    "@context": ["https://www.w3.org/ns/did/v1"],
    "id": "did:key:z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA",
    "verificationMethod": [
        {
            "id": "did:key:z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA#z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA",
            "type": "Bls12381G1Key2020",
            "controller": "did:key:z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA",
            "publicKeyBase58": "6FywSzB5BPd7xehCo1G4nYHAoZPMMP3gd4PLnvgA6SsTsogtz8K7RDznqLpFPLZXAE",
        }
    ],
    "authentication": [
        "did:key:z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA#z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA"
    ],
    "assertionMethod": [
        "did:key:z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA#z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA"
    ],
    "capabilityDelegation": [
        "did:key:z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA#z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA"
    ],
    "capabilityInvocation": [
        "did:key:z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA#z3tEFALUKUzzCAvytMHX8X4SnsNsq6T5tC5Zb18oQEt1FqNcJXqJ3AA9umgzA9yoqPBeWA"
    ],
    "keyAgreement": [],
}

DID_BLS12381G1G2_z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s = {
    "@context": "https://www.w3.org/ns/did/v1",
    "id": "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s",
    "verificationMethod": [
        {
            "id": "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s#z3tEG5qmJZX29jJSX5kyhDR5YJNnefJFdwTxRqk6zbEPv4Pf2xF12BpmXv9NExxSRFGfxd",
            "type": "Bls12381G1Key2020",
            "controller": "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s",
            "publicKeyBase58": "7BVES4h78wzabPAfMhchXyH5d8EX78S5TtzePH2YkftWcE6by9yj3NTAv9nsyCeYch",
        },
        {
            "id": "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s#zUC7LTa4hWtaE9YKyDsMVGiRNqPMN3s4rjBdB3MFi6PcVWReNfR72y3oGW2NhNcaKNVhMobh7aHp8oZB3qdJCs7RebM2xsodrSm8MmePbN25NTGcpjkJMwKbcWfYDX7eHCJjPGM",
            "type": "Bls12381G2Key2020",
            "controller": "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s",
            "publicKeyBase58": "26d2BdqELsXg7ZHCWKL2D5Y2S7mYrpkdhJemSEEvokd4qy4TULJeeU44hYPGKo4x4DbBp5ARzkv1D6xuB3bmhpdpKAXuXtode67wzh9PCtW8kTqQhH19VSiFZkLNkhe9rtf3",
        },
    ],
    "authentication": [
        "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s#z3tEG5qmJZX29jJSX5kyhDR5YJNnefJFdwTxRqk6zbEPv4Pf2xF12BpmXv9NExxSRFGfxd",
        "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s#zUC7LTa4hWtaE9YKyDsMVGiRNqPMN3s4rjBdB3MFi6PcVWReNfR72y3oGW2NhNcaKNVhMobh7aHp8oZB3qdJCs7RebM2xsodrSm8MmePbN25NTGcpjkJMwKbcWfYDX7eHCJjPGM",
    ],
    "assertionMethod": [
        "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s#z3tEG5qmJZX29jJSX5kyhDR5YJNnefJFdwTxRqk6zbEPv4Pf2xF12BpmXv9NExxSRFGfxd",
        "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s#zUC7LTa4hWtaE9YKyDsMVGiRNqPMN3s4rjBdB3MFi6PcVWReNfR72y3oGW2NhNcaKNVhMobh7aHp8oZB3qdJCs7RebM2xsodrSm8MmePbN25NTGcpjkJMwKbcWfYDX7eHCJjPGM",
    ],
    "capabilityDelegation": [
        "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s#z3tEG5qmJZX29jJSX5kyhDR5YJNnefJFdwTxRqk6zbEPv4Pf2xF12BpmXv9NExxSRFGfxd",
        "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s#zUC7LTa4hWtaE9YKyDsMVGiRNqPMN3s4rjBdB3MFi6PcVWReNfR72y3oGW2NhNcaKNVhMobh7aHp8oZB3qdJCs7RebM2xsodrSm8MmePbN25NTGcpjkJMwKbcWfYDX7eHCJjPGM",
    ],
    "capabilityInvocation": [
        "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s#z3tEG5qmJZX29jJSX5kyhDR5YJNnefJFdwTxRqk6zbEPv4Pf2xF12BpmXv9NExxSRFGfxd",
        "did:key:z5TcESXuYUE9aZWYwSdrUEGK1HNQFHyTt4aVpaCTVZcDXQmUheFwfNZmRksaAbBneNm5KyE52SdJeRCN1g6PJmF31GsHWwFiqUDujvasK3wTiDr3vvkYwEJHt7H5RGEKYEp1ErtQtcEBgsgY2DA9JZkHj1J9HZ8MRDTguAhoFtR4aTBQhgnkP4SwVbxDYMEZoF2TMYn3s#zUC7LTa4hWtaE9YKyDsMVGiRNqPMN3s4rjBdB3MFi6PcVWReNfR72y3oGW2NhNcaKNVhMobh7aHp8oZB3qdJCs7RebM2xsodrSm8MmePbN25NTGcpjkJMwKbcWfYDX7eHCJjPGM",
    ],
    "keyAgreement": [],
}



"""Linked data proof verifiable options detail artifacts to attach to RFC 453 messages."""

from typing import Optional, Union

from marshmallow import INCLUDE, fields

from .......messaging.models.base import BaseModel, BaseModelSchema
from .......vc.vc_ld import CredentialSchema
from .......vc.vc_ld.models.credential import VerifiableCredential
from .......vc.vc_ld.models.options import LDProofVCOptions, LDProofVCOptionsSchema


class LDProofVCDetail(BaseModel):
    """Linked data proof verifiable credential detail."""

    class Meta:
        """LDProofVCDetail metadata."""

        schema_class = "LDProofVCDetailSchema"

    def __init__(
        self,
        credential: Optional[Union[dict, VerifiableCredential]],
        options: Optional[Union[dict, LDProofVCOptions]],
    ) -> None:
        """Initialize the LDProofVCDetail instance."""
        self.credential = credential
        self.options = options

    def __eq__(self, other: object) -> bool:
        """Comparison between linked data vc details."""
        if isinstance(other, LDProofVCDetail):
            return self.credential == other.credential and self.options == other.options
        return False


class LDProofVCDetailSchema(BaseModelSchema):
    """Linked data proof verifiable credential detail schema."""

    class Meta:
        """Accept parameter overload."""

        unknown = INCLUDE
        model_class = LDProofVCDetail

    credential = fields.Nested(
        CredentialSchema(),
        required=True,
        metadata={
            "description": "Detail of the JSON-LD Credential to be issued",
            "example": {
                "@context": [
                    "https://www.w3.org/2018/credentials/v1",
                    "https://w3id.org/citizenship/v1",
                ],
                "type": ["VerifiableCredential", "PermanentResidentCard"],
                "issuer": "did:key:z6MkmjY8GnV5i9YTDtPETC2uUAW6ejw3nk5mXF5yci5ab7th",
                "identifier": "83627465",
                "name": "Permanent Resident Card",
                "description": "Government of Example Permanent Resident Card.",
                "issuanceDate": "2019-12-03T12:19:52Z",
                "credentialSubject": {
                    "type": ["PermanentResident", "Person"],
                    "givenName": "JOHN",
                    "familyName": "SMITH",
                    "gender": "Male",
                },
            },
        },
    )

    options = fields.Nested(
        LDProofVCOptionsSchema(),
        required=True,
        metadata={
            "description": (
                "Options for specifying how the linked data proof is created."
            ),
            "example": {"proofType": "Ed25519Signature2018"},
        },
    )



import json
from unittest import IsolatedAsyncioTestCase

from ......connections.models.connection_target import ConnectionTarget
from ......messaging.base_handler import HandlerException
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...messages.forward import Forward
from ...models.route_record import RouteRecord
from .. import forward_handler as test_module

TEST_CONN_ID = "conn-id"
TEST_VERKEY = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
TEST_ROUTE_VERKEY = "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"


class TestForwardHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.context = RequestContext.test_context(await create_test_profile())
        self.context.connection_ready = True
        self.context.message = Forward(to="sample-did", msg={"msg": "sample-message"})

    async def test_handle(self):
        self.context.message_receipt = MessageReceipt(recipient_verkey=TEST_VERKEY)
        handler = test_module.ForwardHandler()

        responder = MockResponder()
        with (
            mock.patch.object(test_module, "RoutingManager", autospec=True) as mock_mgr,
            mock.patch.object(
                test_module, "BaseConnectionManager", autospec=True
            ) as mock_connection_mgr,
            mock.patch.object(
                self.context.profile, "notify", autospec=True
            ) as mock_notify,
        ):
            mock_mgr.return_value.get_recipient = mock.CoroutineMock(
                return_value=RouteRecord(connection_id="dummy")
            )
            mock_connection_mgr.return_value.get_connection_targets = mock.CoroutineMock(
                return_value=[
                    ConnectionTarget(
                        recipient_keys=["recip_key"],
                    )
                ]
            )

            await handler.handle(self.context, responder)

            mock_notify.assert_called_once_with(
                "acapy::forward::received",
                {
                    "connection_id": "dummy",
                    "status": "queued_for_delivery",
                    "recipient_key": "sample-did",
                },
            )

            messages = responder.messages
            assert len(messages) == 1
            (result, target) = messages[0]
            assert json.loads(result) == self.context.message.msg
            assert target["connection_id"] == "dummy"

    async def test_handle_receipt_no_recipient_verkey(self):
        self.context.message_receipt = MessageReceipt()
        handler = test_module.ForwardHandler()
        with self.assertRaises(HandlerException):
            await handler.handle(self.context, None)

    async def test_handle_cannot_resolve_recipient(self):
        self.context.message_receipt = MessageReceipt(recipient_verkey=TEST_VERKEY)
        handler = test_module.ForwardHandler()

        responder = MockResponder()
        with mock.patch.object(test_module, "RoutingManager", autospec=True) as mock_mgr:
            mock_mgr.return_value.get_recipient = mock.CoroutineMock(
                side_effect=test_module.RoutingManagerError()
            )

            await handler.handle(self.context, responder)

            messages = responder.messages
            assert not messages
